%
% STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
% the University of California, through Lawrence Berkeley National Laboratory
% (subject to receipt of any required approvals from the U.S. Dept. of Energy).
% All rights reserved.
%
% If you have questions about your rights to use or distribute this software,
% please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
%
% NOTICE. This software is owned by the U.S. Department of Energy. As such, the
% U.S. Government has been granted for itself and others acting on its behalf a
% paid-up, nonexclusive, irrevocable, worldwide license in the Software to
% reproduce, prepare derivative works, and perform publicly and display publicly.
% Beginning five (5) years after the date permission to assert copyright is
% obtained from the U.S. Department of Energy, and subject to any subsequent five
% (5) year renewals, the U.S. Government is granted for itself and others acting
% on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
% Software to reproduce, prepare derivative works, distribute copies to the
% public, perform publicly and display publicly, and to permit others to do so.
%
% Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
%             (Lawrence Berkeley National Lab, Computational Research Division).
%

% Solution of a linear system with STRUMPACK, Octave/Matlab interface.

n=1024; % Problem size
nrhs=3;

% Toeplitz matrix from Quantum Chemistry
d=0.1;
mypi=3.1416;
r(1)=(mypi)^2/(6*d^2);
for j=2:n
  r(j)=(-1)^(1-j)/((1-j)^2*d^2);
end
A=toeplitz(r);
A=A+1i*A;

% Initialize the SDP structure
sdp=SDP_mex_dcomplex_init();

% Compress the matrix and check the quality
sdp.use_HSS=1;
sdp.levels_HSS=5;
sdp.min_rand_HSS=10;
sdp.lim_rand_HSS=5;
sdp.inc_rand_HSS=10;
sdp.max_rand_HSS=100;
sdp.tol_HSS=1e-12;
sdp.steps_IR=10;
sdp.tol_IR=1e-10;
SDP_mex_dcomplex_compress_A(sdp,A);

% Solve a linear system and check quality
SDP_mex_dcomplex_factor(sdp,A);
rand('seed',0);
B=rand(n,nrhs);
X=SDP_mex_dcomplex_solve(sdp,B);
SDP_mex_dcomplex_check_solution(sdp,A,X,B);
X=SDP_mex_dcomplex_refine(sdp,A,X,B);

% Test the matrix-vector product
R=rand(n,nrhs);
Strue=rand(n,nrhs);
S=Strue;
Strue=2.5*A'*R+1.5*Strue;
S=SDP_mex_dcomplex_product(sdp,'C',2.5,A,R,1.5,S);
if(~sdp.myid)
  fprintf('Product quality: %g\n\n',norm(S-Strue)/norm(Strue));
end

% Test element extraction
I=randi(n,5,1);
J=randi(n,7,1);
Btrue=A(I,J);
B=SDP_mex_dcomplex_extract(sdp,A,I,J);
if(~sdp.myid)
  fprintf('Extraction quality: %g\n\n',norm(B-Btrue,'fro')/norm(Btrue,'fro'));
end

% Print statistics
SDP_mex_dcomplex_print_statistics(sdp);

% Clean up
SDP_mex_dcomplex_destroy(sdp,1);
