/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */

#include "StrumpackDensePackage_mex.hpp"

#define struct_in prhs[0]
#define A_in      prhs[1]
#define X_in      prhs[2]
#define B_in      prhs[3]
#define err_out   plhs[0]

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
  /* err_out=check_solution(struct_in, A_in, X_in, B_in) */
  StrumpackDensePackage<dcomplex,double> *sdp;
  mxArray *field;
  int n, d;
  int np;
  int ctxt;
  int descA[BLACSCTXTSIZE], descXB[BLACSCTXTSIZE];
  long long ptr;
  dcomplex *A, *X, *B, *Acent, *Xcent, *Bcent;
  double *Ar, *Ai, *Xr, *Xi, *Br, *Bi;
  double err;

  if(nrhs!=4) {
    mexErrMsgTxt("Usage: err=SDP_mex_dcomplex_check_solution(sdp,A,X,B)");
  }

  field=mxGetField(struct_in,0,"obj");
  ptr=(long long) mxGetPr(field)[0];
  if(ptr==0) {
    mexErrMsgTxt("Init must be called first!");
  } else {
    sdp=(StrumpackDensePackage<dcomplex,double>*)ptr;
  }

  copystructfields(prhs[0],sdp);

  ctxt=mxGetPr(mxGetField(struct_in,0,"ctxt"))[0];

  n=mxGetM(B_in);
  if(n!=mxGetM(X_in)) {
    mexErrMsgTxt("RHS and solution vectors must have the same row count!");
  }
  if(n!=mxGetM(A_in)) {
    mexErrMsgTxt("RHS and solution vectors must have the same row count as A!");
  }

  d=mxGetN(B_in);
  if(d!=mxGetN(X_in)) {
    mexErrMsgTxt("RHS and solution vectors must have the same column count!");
  }

  Acent=new dcomplex[n*n];
  Ar=mxGetPr(A_in);
  if(mxIsComplex(A_in)) {
    Ai=mxGetPi(A_in);
    for(int i=0;i<n*n;i++)
      Acent[i]=dcomplex(Ar[i],Ai[i]);
  } else {
    for(int i=0;i<n*n;i++)
      Acent[i]=dcomplex(Ar[i],0);
  }

  Xcent=new dcomplex[n*d];
  Xr=mxGetPr(X_in);
  if(mxIsComplex(X_in)) {
    Xi=mxGetPi(X_in);
    for(int i=0;i<n*d;i++)
      Xcent[i]=dcomplex(Xr[i],Xi[i]);
  } else {
    for(int i=0;i<n*d;i++)
      Xcent[i]=dcomplex(Xr[i],0);
  }

  Bcent=new dcomplex[n*d];
  Br=mxGetPr(B_in);
  if(mxIsComplex(B_in)) {
    Bi=mxGetPi(B_in);
    for(int i=0;i<n*d;i++)
      Bcent[i]=dcomplex(Br[i],Bi[i]);
  } else {
    for(int i=0;i<n*d;i++)
      Bcent[i]=dcomplex(Br[i],0);
  }

  MPI_Comm_size(MPI_COMM_WORLD,&np);
  if(np==1) {
    A=Acent;
    X=Xcent;
    B=Bcent;
    descset_(descA,&n,&n,&IONE,&IONE,&IZERO,&IZERO,&INONE,&IONE);
    descset_(descXB,&n,&d,&IONE,&IONE,&IZERO,&IZERO,&INONE,&IONE);
  } else {
    FromCentralizedTo2D(Acent,n,n,&A,ctxt,descA);
    FromCentralizedTo2D(Xcent,n,d,&X,ctxt,descXB);
    FromCentralizedTo2D(Bcent,n,d,&B,ctxt,descXB);
    delete[] Acent;
    delete[] Xcent;
    delete[] Bcent;
  }

  err=sdp->check_solution(A,descA,X,descXB,B,descXB);

  err_out=mxCreateDoubleMatrix(1,1,mxREAL);
  mxGetPr(err_out)[0]=err;

  delete[] A;
  delete[] X;
  delete[] B;

}
