/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */

#include "StrumpackDensePackage_mex.hpp"

#define struct_in prhs[0]
#define A_in      prhs[1]
#define Rr_in     prhs[2]
#define Rc_in     prhs[3]
#define Sr_in     prhs[4]
#define Sc_in     prhs[5]

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
  /* compress(struct_in, A_in, Rr_in, Rc_in, Sr_in, Sc_in) */
  StrumpackDensePackage<dcomplex,double> *sdp;
  mxArray *field;
  int m, n, d;
  int ctxt;
  int np;
  int descA[BLACSCTXTSIZE], descRS[BLACSCTXTSIZE];
  long long ptr;
  double *Ar, *Ai, *Rrr, *Rri, *Rcr, *Rci, *Srr, *Sri, *Scr, *Sci;
  dcomplex *A, *Rr, *Rc, *Sr, *Sc, *Acent, *Rrcent, *Rccent, *Srcent, *Sccent;

  if(nrhs!=6) {
    mexErrMsgTxt("Usage: SDP_mex_dcomplex_compress_ARS(sdp,A,Rr,Rc,Sr,Sc)");
  }

  field=mxGetField(struct_in,0,"obj");
  ptr=(long long) mxGetPr(field)[0];
  if(ptr==0) {
    mexErrMsgTxt("Init must be called first!");
  } else {
    sdp=(StrumpackDensePackage<dcomplex,double>*)ptr;
  }

  copystructfields(prhs[0],sdp);

  ctxt=mxGetPr(mxGetField(struct_in,0,"ctxt"))[0];

  m=mxGetM(A_in);
  n=mxGetN(A_in);

  if(m!=n) {
    mexErrMsgTxt("Matrix must be square!");
  }
  if(m!=mxGetM(Rr_in) || m!=mxGetM(Rc_in) || m!=mxGetM(Rc_in)|| m!=mxGetM(Rc_in)) {
    mexErrMsgTxt("Random vectors and samples must have the same row count as A!");
  }

  d=sdp->min_rand_HSS;
  if(d!=mxGetN(Rr_in) || d!=mxGetN(Rc_in) || d!=mxGetN(Rc_in)|| d!=mxGetN(Rc_in)) {
    mexErrMsgTxt("Random vectors and samples must have column count equal to min_rand_HSS!");
  }

  Acent=new dcomplex[n*n];
  Ar=mxGetPr(A_in);
  if(mxIsComplex(A_in)) {
    Ai=mxGetPi(A_in);
    for(int i=0;i<n*n;i++)
      Acent[i]=dcomplex(Ar[i],Ai[i]);
  } else {
    for(int i=0;i<n*n;i++)
      Acent[i]=dcomplex(Ar[i],0);
  }

  Rrcent=new dcomplex[n*d];
  Rrr=mxGetPr(Rr_in);
  if(mxIsComplex(Rr_in)) {
    Rri=mxGetPi(Rr_in);
    for(int i=0;i<n*d;i++)
      Rrcent[i]=dcomplex(Rrr[i],Rri[i]);
  } else {
    for(int i=0;i<n*d;i++)
      Rrcent[i]=dcomplex(Rrr[i],0);
  }

  Rccent=new dcomplex[n*d];
  Rcr=mxGetPr(Rc_in);
  if(mxIsComplex(Rc_in)) {
    Rci=mxGetPi(Rc_in);
    for(int i=0;i<n*d;i++)
      Rccent[i]=dcomplex(Rcr[i],Rci[i]);
  } else {
    for(int i=0;i<n*d;i++)
      Rccent[i]=dcomplex(Rcr[i],0);
  }

  Srcent=new dcomplex[n*d];
  Srr=mxGetPr(Sr_in);
  if(mxIsComplex(Sr_in)) {
    Sri=mxGetPi(Sr_in);
    for(int i=0;i<n*d;i++)
      Srcent[i]=dcomplex(Srr[i],Sri[i]);
  } else {
    for(int i=0;i<n*d;i++)
      Srcent[i]=dcomplex(Srr[i],0);
  }

  Sccent=new dcomplex[n*d];
  Scr=mxGetPr(Sc_in);
  if(mxIsComplex(Sc_in)) {
    Sci=mxGetPi(Sc_in);
    for(int i=0;i<n*d;i++)
      Sccent[i]=dcomplex(Scr[i],Sci[i]);
  } else {
    for(int i=0;i<n*d;i++)
      Sccent[i]=dcomplex(Scr[i],0);
  }

  MPI_Comm_size(MPI_COMM_WORLD,&np);
  if(np==1) {
    A=Acent;
    Rr=Rrcent;
    Rc=Rccent;
    Sr=Srcent;
    Sc=Sccent;
    descset_(descA,&n,&n,&IONE,&IONE,&IZERO,&IZERO,&INONE,&IONE);
    descset_(descRS,&n,&d,&IONE,&IONE,&IZERO,&IZERO,&INONE,&IONE);
  } else {
    FromCentralizedTo2D(Acent,n,n,&A,ctxt,descA);
    FromCentralizedTo2D(Rrcent,n,d,&Rr,ctxt,descRS);
    FromCentralizedTo2D(Rccent,n,d,&Rc,ctxt,descRS);
    FromCentralizedTo2D(Srcent,n,d,&Sr,ctxt,descRS);
    FromCentralizedTo2D(Sccent,n,d,&Sc,ctxt,descRS);
    delete[] Acent;
    delete[] Rrcent;
    delete[] Rccent;
    delete[] Srcent;
    delete[] Sccent;
  }

  sdp->compress(A,descA,Rr,Rc,Sr,Sc,descRS);

  delete[] A;
  delete[] Rr;
  delete[] Rc;
  delete[] Sr;
  delete[] Sc;

}
