/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */

#include "StrumpackDensePackage_mex.hpp"

#define struct_in prhs[0]
#define X_in      prhs[1]
#define B_in      prhs[2]
#define X_out     plhs[0]

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
  /* X_out=expand_solution(struct_in, X_in, B_in) */
  StrumpackDensePackage<dcomplex,double> *sdp;
  mxArray *field;
  int n, d;
  int ctxt;
  int np;
  int descXB[BLACSCTXTSIZE];
  long long ptr;
  dcomplex *X, *B, *Xcent, *Bcent;
  double *Xr, *Xi, *Br, *Bi;

  if(nrhs!=3) {
    mexErrMsgTxt("Usage: X=SDP_mex_dcomplex_expand_solution(sdp,X,B)");
  }

  field=mxGetField(struct_in,0,"obj");
  ptr=(long long) mxGetPr(field)[0];
  if(ptr==0) {
    mexErrMsgTxt("Init must be called first!");
  } else {
    sdp=(StrumpackDensePackage<dcomplex,double>*)ptr;
  }

  copystructfields(prhs[0],sdp);

  ctxt=mxGetPr(mxGetField(struct_in,0,"ctxt"))[0];

  n=mxGetM(B_in);
  d=mxGetN(B_in);

  Bcent=new dcomplex[n*d];
  Br=mxGetPr(B_in);
  if(mxIsComplex(B_in)) {
    Bi=mxGetPi(B_in);
    for(int i=0;i<n*d;i++)
      Bcent[i]=dcomplex(Br[i],Bi[i]);
  } else {
    for(int i=0;i<n*d;i++)
      Bcent[i]=dcomplex(Br[i],0);
  }

  Xcent=new dcomplex[n*d];
  Xr=mxGetPr(X_in);
  if(mxIsComplex(X_in)) {
    Xi=mxGetPi(X_in);
    for(int i=0;i<n*d;i++)
      Xcent[i]=dcomplex(Xr[i],Xi[i]);
  } else {
    for(int i=0;i<n*d;i++)
      Xcent[i]=dcomplex(Xr[i],0);
  }

  MPI_Comm_size(MPI_COMM_WORLD,&np);
  if(np==1) {
    X=Xcent;
    B=Bcent;
    descset_(descXB,&n,&d,&IONE,&IONE,&IZERO,&IZERO,&INONE,&IONE);
  } else {
    FromCentralizedTo2D(Xcent,n,d,&X,ctxt,descXB);
    FromCentralizedTo2D(Bcent,n,d,&B,ctxt,descXB);
    delete[] Xcent;
    delete[] Bcent;
  }

  sdp->expand_solution(X,descXB,B,descXB);

  delete[] B;

  if(np==1) {
    Xcent=X;
  } else {
    From2DToReplicated(X,n,d,ctxt,&Xcent);
    delete[] X;
  }

  X_out=mxCreateDoubleMatrix(n,d,mxCOMPLEX);
  Xr=mxGetPr(X_out);
  Xi=mxGetPi(X_out);
  for(int i=0;i<n*d;i++) {
    Xr[i]=Xcent[i].real();
    Xi[i]=Xcent[i].imag();
  }

  delete[] Xcent;

}
