/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */

#include "StrumpackDensePackage_mex.hpp"

#define struct_in prhs[0]
#define A_in      prhs[1]

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
  /* factor(struct_in, A_in) */
  StrumpackDensePackage<dcomplex,double> *sdp;
  mxArray *field;
  int m, n;
  int ctxt;
  int np;
  int descA[BLACSCTXTSIZE];
  long long ptr;
  dcomplex *A, *Acent;
  double *Ar, *Ai;

  if(nrhs!=2) {
    mexErrMsgTxt("Usage: SDP_mex_dcomplex_factor(sdp,A)");
  }

  field=mxGetField(struct_in,0,"obj");
  ptr=(long long) mxGetPr(field)[0];
  if(ptr==0) {
    mexErrMsgTxt("Init must be called first!");
  } else {
    sdp=(StrumpackDensePackage<dcomplex,double>*)ptr;
  }

  copystructfields(prhs[0],sdp);

  ctxt=mxGetPr(mxGetField(struct_in,0,"ctxt"))[0];

  m=mxGetM(A_in);
  n=mxGetN(A_in);

  if(m!=n) {
    mexErrMsgTxt("Matrix must be square!");
  }

  Acent=new dcomplex[n*n];
  Ar=mxGetPr(A_in);
  if(mxIsComplex(A_in)) {
    Ai=mxGetPi(A_in);
    for(int i=0;i<n*n;i++)
      Acent[i]=dcomplex(Ar[i],Ai[i]);
  } else {
    for(int i=0;i<n*n;i++)
      Acent[i]=dcomplex(Ar[i],0);
  }

  MPI_Comm_size(MPI_COMM_WORLD,&np);
  if(np==1) {
    A=Acent;
    descset_(descA,&n,&n,&IONE,&IONE,&IZERO,&IZERO,&INONE,&IONE);
  } else {
    FromCentralizedTo2D(Acent,n,n,&A,ctxt,descA);
    delete[] Acent;
  }

  sdp->factor(A,descA);

  delete[] A;

}
