/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */

#include "StrumpackDensePackage_mex.hpp"

#define struct_in prhs[0]
#define Trans_in  prhs[1]
#define alpha_in  prhs[2]
#define B_in      prhs[3]
#define beta_in   prhs[4]
#define C_in      prhs[5]
#define C_out     plhs[0]

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
  /* C_out=schur_product(struct_in, Trans_in, alpha_in, B_in, beta_in, C_in) */
  StrumpackDensePackage<dcomplex,double> *sdp;
  mxArray *field;
  int n, d;
  int ctxt;
  int np;
  int descBC[BLACSCTXTSIZE];
  long long ptr;
  dcomplex *B, *C, *Bcent, *Ccent;
  dcomplex alpha, beta;
  char Trans;
  double *Br, *Bi, *Cr, *Ci;

  if(nrhs!=6) {
    mexErrMsgTxt("Usage: C=SDP_mex_double_schur_product(sdp,Trans,alpha,B,beta,C)");
  }

  field=mxGetField(struct_in,0,"obj");
  ptr=(long long) mxGetPr(field)[0];
  if(ptr==0) {
    mexErrMsgTxt("Init must be called first!");
  } else {
    sdp=(StrumpackDensePackage<dcomplex,double>*)ptr;
  }

  copystructfields(prhs[0],sdp);

  ctxt=mxGetPr(mxGetField(struct_in,0,"ctxt"))[0];

  n=mxGetM(B_in);
  if(n!=mxGetM(C_in)) {
    mexErrMsgTxt("Vectors must have the same row count!");
  }

  d=mxGetN(B_in);
  if(d!=mxGetN(C_in)) {
    mexErrMsgTxt("Vectors must have the same column count!");
  }

  Trans=mxGetChars(Trans_in)[0];

  if(mxIsComplex(alpha_in))
    alpha=dcomplex(mxGetPr(alpha_in)[0],mxGetPi(alpha_in)[0]);
  else
    alpha=mxGetPr(alpha_in)[0];

  if(mxIsComplex(beta_in))
    beta=dcomplex(mxGetPr(beta_in)[0],mxGetPi(beta_in)[0]);
  else
    beta=mxGetPr(beta_in)[0];

  Bcent=new dcomplex[n*d];
  Br=mxGetPr(B_in);
  if(mxIsComplex(B_in)) {
    Bi=mxGetPi(B_in);
    for(int i=0;i<n*d;i++)
      Bcent[i]=dcomplex(Br[i],Bi[i]);
  } else {
    for(int i=0;i<n*d;i++)
      Bcent[i]=dcomplex(Br[i],0);
  }

  Ccent=new dcomplex[n*d];
  Cr=mxGetPr(C_in);
  if(mxIsComplex(C_in)) {
    Ci=mxGetPi(C_in);
    for(int i=0;i<n*d;i++)
      Ccent[i]=dcomplex(Cr[i],Ci[i]);
  } else {
    for(int i=0;i<n*d;i++)
      Ccent[i]=dcomplex(Cr[i],0);
  }

  MPI_Comm_size(MPI_COMM_WORLD,&np);
  if(np==1) {
    B=Bcent;
    C=Ccent;
    descset_(descBC,&n,&d,&IONE,&IONE,&IZERO,&IZERO,&INONE,&IONE);
  } else {
    FromCentralizedTo2D(Bcent,n,d,&B,ctxt,descBC);
    FromCentralizedTo2D(Ccent,n,d,&C,ctxt,descBC);
    delete[] Bcent;
    delete[] Ccent;
  }

  sdp->schur_product(Trans,alpha,B,descBC,beta,C,descBC);

  delete[] B;

  if(np==1) {
    Ccent=C;
  } else {
    From2DToReplicated(C,n,d,ctxt,&Ccent);
    delete[] C;
  }

  C_out=mxCreateDoubleMatrix(n,d,mxCOMPLEX);
  Cr=mxGetPr(C_out);
  Ci=mxGetPi(C_out);
  for(int i=0;i<n*d;i++) {
    Cr[i]=Ccent[i].real();
    Ci[i]=Ccent[i].imag();
  }

  delete[] Ccent;
}
