/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */

#include "StrumpackDensePackage_mex.hpp"

#define struct_in prhs[0]
#define A_in      prhs[1]
#define I_in      prhs[2]
#define J_in      prhs[3]
#define B_out     plhs[0]

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
  /* B_out=extract(struct_in, A_in, I_in, J_in) */
  StrumpackDensePackage<double,double> *sdp;
  mxArray *field;
  int n, nI, nJ;
  int ctxt;
  int np;
  int descA[BLACSCTXTSIZE], descB[BLACSCTXTSIZE];
  long long ptr;
  double *A, *B, *Bcent;
  double *Br;
  int *I, *J;

  if(nrhs!=4) {
    mexErrMsgTxt("Usage: B=SDP_mex_double_extract(sdp,A,I,J)");
  }

  field=mxGetField(struct_in,0,"obj");
  ptr=(long long) mxGetPr(field)[0];
  if(ptr==0) {
    mexErrMsgTxt("Init must be called first!");
  } else {
    sdp=(StrumpackDensePackage<double,double>*)ptr;
  }

  copystructfields(prhs[0],sdp);

  ctxt=mxGetPr(mxGetField(struct_in,0,"ctxt"))[0];

  nI=mxGetM(I_in);
  if(nI==1)
    nI=mxGetN(I_in);

  nJ=mxGetM(J_in);
  if(nJ==1)
    nJ=mxGetN(J_in);

  I=new int[nI];
  for(int i=0;i<nI;i++)
    I[i]=mxGetPr(I_in)[i];

  J=new int[nJ];
  for(int j=0;j<nJ;j++)
    J[j]=mxGetPr(J_in)[j];

  n=mxGetM(A_in);

  A=mxGetPr(A_in);

  B=NULL;

  MPI_Comm_size(MPI_COMM_WORLD,&np);
  if(np==1) {
    B=new double[nI*nJ];
    descset_(descA,&n,&n,&IONE,&IONE,&IZERO,&IZERO,&INONE,&IONE);
    descset_(descB,&nI,&nJ,&IONE,&IONE,&IZERO,&IZERO,&INONE,&IONE);
  } else {
    FromCentralizedTo2D(A,n,n,&A,ctxt,descA);
    FromCentralizedTo2D(B,nI,nJ,&B,ctxt,descB);
  }

  sdp->extract(A,descA,B,descB,I,nI,J,nJ);

  if(np==1) {
    Bcent=B;
  } else {
    From2DToReplicated(B,nI,nJ,ctxt,&Bcent);
    delete[] B;
  }

  B_out=mxCreateDoubleMatrix(nI,nJ,mxREAL);
  Br=mxGetPr(B_out);
  for(int i=0;i<nI*nJ;i++) {
    Br[i]=Bcent[i];
  }

  delete[] I;
  delete[] J;
  delete[] Bcent;

}
