/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */

#include "StrumpackDensePackage_mex.hpp"

#define ptr_out plhs[0]

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
  /* ptr_out=init() (calls the C++ constructor) */
  StrumpackDensePackage<double,double> *sdp;
  mxArray *field;
  double *pr;
  int flag;
  int ctxt, nprow, npcol, myid, np;

  if(nrhs!=0 || nlhs!=1) {
    mexErrMsgTxt("Usage: sdp=SDP_mex_double_init()");
  }

  MPI_Initialized(&flag);
  if(!flag)
    MPI_Init(NULL,NULL);

  /* Initialize the SDP C structure */
  sdp=new StrumpackDensePackage<double,double>(MPI_COMM_WORLD);

  /* Initialize the SDP Matlab structure */
  ptr_out=mxCreateStructMatrix(1,1,fieldNum,fieldNames);

  /* Copy fields one by one */
  field=mxCreateDoubleMatrix(1,1,mxREAL);
  pr=mxGetPr(field);
  pr[0]=sdp->use_HSS;
  mxSetField(ptr_out,0,"use_HSS",field);

  field=mxCreateDoubleMatrix(1,1,mxREAL);
  pr=mxGetPr(field);
  pr[0]=sdp->tol_HSS;
  mxSetField(ptr_out,0,"tol_HSS",field);

  field=mxCreateDoubleMatrix(1,1,mxREAL);
  pr=mxGetPr(field);
  pr[0]=sdp->min_rand_HSS;
  mxSetField(ptr_out,0,"min_rand_HSS",field);

  field=mxCreateDoubleMatrix(1,1,mxREAL);
  pr=mxGetPr(field);
  pr[0]=sdp->max_rand_HSS;
  mxSetField(ptr_out,0,"max_rand_HSS",field);

  field=mxCreateDoubleMatrix(1,1,mxREAL);
  pr=mxGetPr(field);
  pr[0]=sdp->lim_rand_HSS;
  mxSetField(ptr_out,0,"lim_rand_HSS",field);

  field=mxCreateDoubleMatrix(1,1,mxREAL);
  pr=mxGetPr(field);
  pr[0]=sdp->inc_rand_HSS;
  mxSetField(ptr_out,0,"inc_rand_HSS",field);

  field=mxCreateDoubleMatrix(1,1,mxREAL);
  pr=mxGetPr(field);
  pr[0]=sdp->block_HSS;
  mxSetField(ptr_out,0,"block_HSS",field);

  field=mxCreateDoubleMatrix(1,1,mxREAL);
  pr=mxGetPr(field);
  pr[0]=sdp->levels_HSS;
  mxSetField(ptr_out,0,"levels_HSS",field);

  field=mxCreateDoubleMatrix(1,1,mxREAL);
  pr=mxGetPr(field);
  pr[0]=sdp->split_HSS;
  mxSetField(ptr_out,0,"split_HSS",field);

  field=mxCreateDoubleMatrix(1,1,mxREAL);
  pr=mxGetPr(field);
  pr[0]=sdp->tol_IR;
  mxSetField(ptr_out,0,"tol_IR",field);

  field=mxCreateDoubleMatrix(1,1,mxREAL);
  pr=mxGetPr(field);
  pr[0]=sdp->steps_IR;
  mxSetField(ptr_out,0,"steps_IR",field);

  field=mxCreateDoubleMatrix(1,1,mxREAL);
  pr=mxGetPr(field);
  pr[0]=sdp->fast_IR;
  mxSetField(ptr_out,0,"fast_IR",field);

  field=mxCreateDoubleMatrix(1,1,mxREAL);
  pr=mxGetPr(field);
  pr[0]=sdp->verbose;
  mxSetField(ptr_out,0,"verbose",field);

  field=mxCreateDoubleMatrix(1,1,mxREAL);
  pr=mxGetPr(field);
  pr[0]=sdp->access_is_global;
  mxSetField(ptr_out,0,"access_is_global",field);

  field=mxCreateDoubleMatrix(1,1,mxREAL);
  pr=mxGetPr(field);
  pr[0]=sdp->sched_strat;
  mxSetField(ptr_out,0,"sched_strat",field);

  /* The obj field is special: it is the pointer to the C structure
   * (not the pointer to the C++ object that is stored in sdp->obj).
   */
  field=mxCreateDoubleMatrix(1,1,mxREAL);
  pr=mxGetPr(field);
  pr[0]=(long long) sdp;
  mxSetField(ptr_out,0,"obj",field);

  /* The ctxt field is a BLACS grid that we use for all the
   * matrices we pass to the SDP instance.
   */
  MPI_Comm_size(MPI_COMM_WORLD,&np);
  nprow=floor(sqrt((float)np));
  npcol=np/nprow;
  blacs_get_(&IZERO,&IZERO,&ctxt);
  blacs_gridinit_(&ctxt,"R",&nprow,&npcol);
  field=mxCreateDoubleMatrix(1,1,mxREAL);
  pr=mxGetPr(field);
  pr[0]=ctxt;
  mxSetField(ptr_out,0,"ctxt",field);

  /* Information fields */
  MPI_Comm_rank(MPI_COMM_WORLD,&myid);
  field=mxCreateDoubleMatrix(1,1,mxREAL);
  pr=mxGetPr(field);
  pr[0]=myid;
  mxSetField(ptr_out,0,"myid",field);

  field=mxCreateDoubleMatrix(1,1,mxREAL);
  pr=mxGetPr(field);
  pr[0]=np;
  mxSetField(ptr_out,0,"np",field);

  field=mxCreateDoubleMatrix(1,1,mxREAL);
  pr=mxGetPr(field);
  pr[0]=nprow;
  mxSetField(ptr_out,0,"nprow",field);

  field=mxCreateDoubleMatrix(1,1,mxREAL);
  pr=mxGetPr(field);
  pr[0]=npcol;
  mxSetField(ptr_out,0,"npcol",field);

}
