!
! STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
! the University of California, through Lawrence Berkeley National Laboratory
! (subject to receipt of any required approvals from the U.S. Dept. of Energy).
! All rights reserved.
!
! If you have questions about your rights to use or distribute this software,
! please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
!
! NOTICE. This software is owned by the U.S. Department of Energy. As such, the
! U.S. Government has been granted for itself and others acting on its behalf a
! paid-up, nonexclusive, irrevocable, worldwide license in the Software to
! reproduce, prepare derivative works, and perform publicly and display publicly.
! Beginning five (5) years after the date permission to assert copyright is
! obtained from the U.S. Department of Energy, and subject to any subsequent five
! (5) year renewals, the U.S. Government is granted for itself and others acting
! on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
! Software to reproduce, prepare derivative works, distribute copies to the
! public, perform publicly and display publicly, and to permit others to do so.
!
! Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
!             (Lawrence Berkeley National Lab, Computational Research Division).
!

module StrumpackDensePackage
  use iso_c_binding

  implicit none

  ! Structures StrumpackDensePackage_F90_* mimic the StrumpackDensePackage_C structs.
  ! We explicitly define four versions: double, float, dcomplex, scomplex.
  ! We could use macros but it would make using different arithmetic
  ! simultaneously complicated.
  !
  ! The Fortran functions call the functions in the C interface using ISO C bindings.
  !

  !----------------------------------------------------------------------------
  ! Real double precision
  type, &
  bind(C) :: StrumpackDensePackage_F90_double
    ! User options
    integer(C_INT) :: use_HSS          ! Use HSS compression/factorization/solve or classical LU?
    real(C_DOUBLE) :: tol_HSS          ! Tolerance for HSS compression
    integer(C_INT) :: min_rand_HSS     ! Minimum number of random vectors for HSS compression
    integer(C_INT) :: max_rand_HSS     ! Maximum number of random vectors for HSS compression
    integer(C_INT) :: lim_rand_HSS     ! Criterion of increasing the number of random vectors
    integer(C_INT) :: inc_rand_HSS     ! Number of extra random vectors at each increase
    integer(C_INT) :: block_HSS        ! Block size for the leaves of the HSS tree
    integer(C_INT) :: levels_HSS       ! Number of levels in the HSS tree
    integer(C_INT) :: split_HSS        ! Constrained partitioning of the root node
    real(C_DOUBLE) :: tol_IR           ! Tolerance for iterative refinement
    integer(C_INT) :: steps_IR         ! Number of steps of iterative refinement
    integer(C_INT) :: fast_IR          ! Matvecs with HSS(A) or A in iterative refinement?
    integer(C_INT) :: verbose          ! Verbose or silent?
    integer(C_INT) :: access_is_global ! User-given access routine: requires global synchronization?
    integer(C_INT) :: sched_strat      ! Scheduling strategy: 1 or 2
    type(C_PTR)    :: obj              ! Handle to user data/object, for strategy 2

    ! Handle to the C++ class
    type(C_PTR) :: SDP_CPP
  end type StrumpackDensePackage_F90_double

  !----------------------------------------------------------------------------
  ! Real single precision
  type, &
  bind(C) :: StrumpackDensePackage_F90_float
    ! User options
    integer(C_INT) :: use_HSS          ! Use HSS compression/factorization/solve or classical LU?
    real(C_FLOAT)  :: tol_HSS          ! Tolerance for HSS compression
    integer(C_INT) :: min_rand_HSS     ! Minimum number of random vectors for HSS compression
    integer(C_INT) :: max_rand_HSS     ! Maximum number of random vectors for HSS compression
    integer(C_INT) :: lim_rand_HSS     ! Criterion of increasing the number of random vectors
    integer(C_INT) :: inc_rand_HSS     ! Number of extra random vectors at each increase
    integer(C_INT) :: block_HSS        ! Block size for the leaves of the HSS tree
    integer(C_INT) :: levels_HSS       ! Number of levels in the HSS tree
    integer(C_INT) :: split_HSS        ! Constrained partitioning of the root node
    real(C_FLOAT)  :: tol_IR           ! Tolerance for iterative refinement
    integer(C_INT) :: steps_IR         ! Number of steps of iterative refinement
    integer(C_INT) :: fast_IR          ! Matvecs with HSS(A) or A in iterative refinement?
    integer(C_INT) :: verbose          ! Verbose or silent?
    integer(C_INT) :: access_is_global ! User-given access routine: requires global synchronization?
    integer(C_INT) :: sched_strat      ! Scheduling strategy: 1 or 2
    type(C_PTR)    :: obj              ! Handle to user data/object, for strategy 2

    ! Handle to the C++ class
    type(C_PTR) :: SDP_CPP
  end type StrumpackDensePackage_F90_float

  !----------------------------------------------------------------------------
  ! Complex double precision
  type, &
  bind(C) :: StrumpackDensePackage_F90_dcomplex
    ! User options
    integer(C_INT) :: use_HSS          ! Use HSS compression/factorization/solve or classical LU?
    real(C_DOUBLE) :: tol_HSS          ! Tolerance for HSS compression
    integer(C_INT) :: min_rand_HSS     ! Minimum number of random vectors for HSS compression
    integer(C_INT) :: max_rand_HSS     ! Maximum number of random vectors for HSS compression
    integer(C_INT) :: lim_rand_HSS     ! Criterion of increasing the number of random vectors
    integer(C_INT) :: inc_rand_HSS     ! Number of extra random vectors at each increase
    integer(C_INT) :: block_HSS        ! Block size for the leaves of the HSS tree
    integer(C_INT) :: levels_HSS       ! Number of levels in the HSS tree
    integer(C_INT) :: split_HSS        ! Constrained partitioning of the root node
    real(C_DOUBLE) :: tol_IR           ! Tolerance for iterative refinement
    integer(C_INT) :: steps_IR         ! Number of steps of iterative refinement
    integer(C_INT) :: fast_IR          ! Matvecs with HSS(A) or A in iterative refinement?
    integer(C_INT) :: verbose          ! Verbose or silent?
    integer(C_INT) :: access_is_global ! User-given access routine: requires global synchronization?
    integer(C_INT) :: sched_strat      ! Scheduling strategy: 1 or 2
    type(C_PTR)    :: obj              ! Handle to user data/object, for strategy 2

    ! Handle to the C++ class
    type(C_PTR) :: SDP_CPP
  end type StrumpackDensePackage_F90_dcomplex

  !----------------------------------------------------------------------------
  ! Complex single precision
  type, &
  bind(C) :: StrumpackDensePackage_F90_scomplex
    ! User options
    integer(C_INT) :: use_HSS          ! Use HSS compression/factorization/solve or classical LU?
    real(C_FLOAT)  :: tol_HSS          ! Tolerance for HSS compression
    integer(C_INT) :: min_rand_HSS     ! Minimum number of random vectors for HSS compression
    integer(C_INT) :: max_rand_HSS     ! Maximum number of random vectors for HSS compression
    integer(C_INT) :: lim_rand_HSS     ! Criterion of increasing the number of random vectors
    integer(C_INT) :: inc_rand_HSS     ! Number of extra random vectors at each increase
    integer(C_INT) :: block_HSS        ! Block size for the leaves of the HSS tree
    integer(C_INT) :: levels_HSS       ! Number of levels in the HSS tree
    integer(C_INT) :: split_HSS        ! Constrained partitioning of the root node
    real(C_FLOAT)  :: tol_IR           ! Tolerance for iterative refinement
    integer(C_INT) :: steps_IR         ! Number of steps of iterative refinement
    integer(C_INT) :: fast_IR          ! Matvecs with HSS(A) or A in iterative refinement?
    integer(C_INT) :: verbose          ! Verbose or silent?
    integer(C_INT) :: access_is_global ! User-given access routine: requires global synchronization?
    integer(C_INT) :: sched_strat      ! Scheduling strategy: 1 or 2
    type(C_PTR)    :: obj              ! Handle to user data/object, for strategy 2

    ! Handle to the C++ class
    type(C_PTR) :: SDP_CPP
  end type StrumpackDensePackage_F90_scomplex

  integer, parameter :: BLACSCTXTSIZE=9
  integer, parameter :: INONE=-1
  integer, parameter :: IZERO= 0
  integer, parameter :: IONE = 1

  integer :: indxl2g, numroc
  external :: indxl2g, numroc

interface

  !----------------------------------------------------------------------------
  ! Real double precision

  subroutine SDP_F90_double_init(SDP_F90,i) &
  bind (C,name='SDP_C_double_init_F90')
    ! Initialization
    ! Here we don't call the C "init function" because MPI communicators do not
    ! have the same type in C (MPI_Comm) and Fortran (integer).
    import :: StrumpackDensePackage_F90_double
    import :: C_INT
    implicit none
    type(StrumpackDensePackage_F90_double) :: SDP_F90
    integer(C_INT), value :: i
  end subroutine SDP_F90_double_init

  subroutine SDP_F90_double_destroy(SDP_F90) &
  bind (C,name='SDP_C_double_destroy')
    ! Cleanup
    import :: StrumpackDensePackage_F90_double
    implicit none
    type(StrumpackDensePackage_F90_double) :: SDP_F90
  end subroutine SDP_F90_double_destroy

  subroutine SDP_F90_double_compress_A(SDP_F90,A,descA) &
  bind (C,name='SDP_C_double_compress_A')
    ! Compression with an explicit matrix
    import :: StrumpackDensePackage_F90_double
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_double) :: SDP_F90
    type(C_PTR), value :: A
    type(C_PTR), value :: descA
  end subroutine SDP_F90_double_compress_A

  subroutine SDP_F90_double_compress_ARS(SDP_F90,A,descA,Rr,Rc,Sr,Sc,descRS) &
  bind (C,name='SDP_C_double_compress_ARS')
    ! Compression with an explicit matrix and user-given samples and random vectors
    import :: StrumpackDensePackage_F90_double
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_double) :: SDP_F90
    type(C_PTR), value :: A
    type(C_PTR), value :: descA
    type(C_PTR), value :: Rr
    type(C_PTR), value :: Rc
    type(C_PTR), value :: Sr
    type(C_PTR), value :: Sc
    type(C_PTR), value :: descRS
  end subroutine SDP_F90_double_compress_ARS

  subroutine SDP_F90_double_compress_mtxfree(SDP_F90,Rr,Rc,Sr,Sc,descRS,elements) &
  bind (C,name='SDP_C_double_compress_mtxfree')
    ! Matrix-free compression
    import :: StrumpackDensePackage_F90_double
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_double) :: SDP_F90
    type(C_PTR), value :: Rr
    type(C_PTR), value :: Rc
    type(C_PTR), value :: Sr
    type(C_PTR), value :: Sc
    type(C_PTR), value :: descRS
    type(C_PTR), value :: elements
  end subroutine SDP_F90_double_compress_mtxfree

  real(C_DOUBLE) function SDP_F90_double_check_compression(SDP_F90,A,descA) &
  bind (C,name='SDP_C_double_check_compression')
    ! Compression quality check
    import :: StrumpackDensePackage_F90_double
    import :: C_DOUBLE
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_double) :: SDP_F90
    type(C_PTR), value :: A
    type(C_PTR), value :: descA
  end function SDP_F90_double_check_compression

  subroutine SDP_F90_double_factor(SDP_F90,A,descA) &
  bind (C,name='SDP_C_double_factor')
    ! Factorization
    import :: StrumpackDensePackage_F90_double
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_double) :: SDP_F90
    type(C_PTR), value :: A
    type(C_PTR), value :: descA
  end subroutine SDP_F90_double_factor

  subroutine SDP_F90_double_partially_factor(SDP_F90,A,descA,nfact) &
  bind (C,name='SDP_C_double_partially_factor')
    ! Partial factorization
    import :: StrumpackDensePackage_F90_double
    import :: C_PTR
    import :: C_INT
    implicit none
    type(StrumpackDensePackage_F90_double) :: SDP_F90
    type(C_PTR), value    :: A
    type(C_PTR), value    :: descA
    integer(C_INT), value :: nfact
  end subroutine SDP_F90_double_partially_factor

  subroutine SDP_F90_double_compute_schur(SDP_F90) &
  bind (C,name='SDP_C_double_compute_schur')
    ! Schur complement computation
    import :: StrumpackDensePackage_F90_double
    implicit none
    type(StrumpackDensePackage_F90_double) :: SDP_F90
  end subroutine SDP_F90_double_compute_schur

  subroutine SDP_F90_double_reduce_RHS(SDP_F90,X,descX,B,descB) &
  bind (C,name='SDP_C_double_reduce_RHS')
    ! RHS reduction/condensation
    import :: StrumpackDensePackage_F90_double
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_double) :: SDP_F90
    type(C_PTR), value :: X
    type(C_PTR), value :: descX
    type(C_PTR), value :: B
    type(C_PTR), value :: descB
  end subroutine SDP_F90_double_reduce_RHS

  subroutine SDP_F90_double_schur_product(SDP_F90,Trans,alpha,X,descX,beta,B,descB) &
  bind (C,name='SDP_C_double_schur_product')
    ! Schur complement product
    import :: StrumpackDensePackage_F90_double
    import :: C_CHAR
    import :: C_DOUBLE
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_double) :: SDP_F90
    character(C_CHAR), value :: Trans
    real(C_DOUBLE), value    :: alpha
    type(C_PTR), value       :: X
    type(C_PTR), value       :: descX
    real(C_DOUBLE), value    :: beta
    type(C_PTR), value       :: B
    type(C_PTR), value       :: descB
  end subroutine SDP_F90_double_schur_product

  subroutine SDP_F90_double_expand_solution(SDP_F90,X,descX,B,descB) &
  bind (C,name='SDP_C_double_expand_solution')
    ! Solution expansion
    import :: StrumpackDensePackage_F90_double
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_double) :: SDP_F90
    type(C_PTR), value :: X
    type(C_PTR), value :: descX
    type(C_PTR), value :: B
    type(C_PTR), value :: descB
  end subroutine SDP_F90_double_expand_solution

  subroutine SDP_F90_double_solve(SDP_F90,X,descX,B,descB) &
  bind (C,name='SDP_C_double_solve')
    ! Triangular solve
    import :: StrumpackDensePackage_F90_double
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_double) :: SDP_F90
    type(C_PTR), value :: X
    type(C_PTR), value :: descX
    type(C_PTR), value :: B
    type(C_PTR), value :: descB
  end subroutine SDP_F90_double_solve

  real(C_DOUBLE) function SDP_F90_double_check_solution(SDP_F90,A,descA,X,descX,B,descB) &
  bind (C,name='SDP_C_double_check_solution')
    ! Solution accuracy check
    import :: StrumpackDensePackage_F90_double
    import :: C_DOUBLE
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_double) :: SDP_F90
    type(C_PTR), value :: A
    type(C_PTR), value :: descA
    type(C_PTR), value :: X
    type(C_PTR), value :: descX
    type(C_PTR), value :: B
    type(C_PTR), value :: descB
  end function SDP_F90_double_check_solution

  real(C_DOUBLE) function SDP_F90_double_refine(SDP_F90,A,descA,X,descX,B,descB) &
  bind (C,name='SDP_C_double_refine')
    ! Iterative refinement
    import :: StrumpackDensePackage_F90_double
    import :: C_DOUBLE
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_double) :: SDP_F90
    type(C_PTR), value :: A
    type(C_PTR), value :: descA
    type(C_PTR), value :: X
    type(C_PTR), value :: descX
    type(C_PTR), value :: B
    type(C_PTR), value :: descB
  end function SDP_F90_double_refine

  subroutine SDP_F90_double_product(SDP_F90,Trans,alpha,A,descA,B,descB,beta,C,descC) &
  bind (C,name='SDP_C_double_product')
    ! Matrix-vector product
    import :: StrumpackDensePackage_F90_double
    import :: C_CHAR
    import :: C_DOUBLE
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_double) :: SDP_F90
    character(C_CHAR), value :: Trans
    real(C_DOUBLE), value    :: alpha
    type(C_PTR), value       :: A
    type(C_PTR), value       :: descA
    type(C_PTR), value       :: B
    type(C_PTR), value       :: descB
    real(C_DOUBLE), value    :: beta
    type(C_PTR), value       :: C
    type(C_PTR), value       :: descC
  end subroutine SDP_F90_double_product

  subroutine SDP_F90_double_extract(SDP_F90,A,descA,B,descB,I,nI,J,nJ) &
  bind (C,name='SDP_C_double_extract')
    ! Element extraction
    import :: StrumpackDensePackage_F90_double
    import :: C_PTR
    import :: C_INT
    implicit none
    type(StrumpackDensePackage_F90_double) :: SDP_F90
    type(C_PTR), value    :: A
    type(C_PTR), value    :: descA
    type(C_PTR), value    :: B
    type(C_PTR), value    :: descB
    type(C_PTR), value    :: I
    integer(C_INT), value :: nI
    type(C_PTR), value    :: J
    integer(C_INT), value :: nJ
  end subroutine SDP_F90_double_extract

  subroutine SDP_F90_double_extract_schur(SDP_F90,B,descB,I,nI,J,nJ) &
  bind (C,name='SDP_C_double_extract_schur')
    ! Schur complement element extraction
    import :: StrumpackDensePackage_F90_double
    import :: C_PTR
    import :: C_INT
    implicit none
    type(StrumpackDensePackage_F90_double) :: SDP_F90
    type(C_PTR), value    :: B
    type(C_PTR), value    :: descB
    type(C_PTR), value    :: I
    integer(C_INT), value :: nI
    type(C_PTR), value    :: J
    integer(C_INT), value :: nJ
  end subroutine SDP_F90_double_extract_schur

  subroutine SDP_F90_double_print_statistics(SDP_F90) &
  bind (C,name='SDP_C_double_print_statistics')
    ! Statistics
    import :: StrumpackDensePackage_F90_double
    type(StrumpackDensePackage_F90_double) :: SDP_F90
  end subroutine SDP_F90_double_print_statistics

  !----------------------------------------------------------------------------
  ! Real single precision

  subroutine SDP_F90_float_init(SDP_F90,i) &
  bind (C,name='SDP_C_float_init_F90')
    ! Initialization
    ! Here we don't call the C "init function" because MPI communicators do not
    ! have the same type in C (MPI_Comm) and Fortran (integer).
    import :: StrumpackDensePackage_F90_float
    import :: C_INT
    implicit none
    type(StrumpackDensePackage_F90_float) :: SDP_F90
    integer(C_INT), value :: i
  end subroutine SDP_F90_float_init

  subroutine SDP_F90_float_destroy(SDP_F90) &
  bind (C,name='SDP_C_float_destroy')
    ! Cleanup
    import :: StrumpackDensePackage_F90_float
    implicit none
    type(StrumpackDensePackage_F90_float) :: SDP_F90
  end subroutine SDP_F90_float_destroy

  subroutine SDP_F90_float_compress_A(SDP_F90,A,descA) &
  bind (C,name='SDP_C_float_compress_A')
    ! Compression with an explicit matrix
    import :: StrumpackDensePackage_F90_float
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_float) :: SDP_F90
    type(C_PTR), value :: A
    type(C_PTR), value :: descA
  end subroutine SDP_F90_float_compress_A

  subroutine SDP_F90_float_compress_ARS(SDP_F90,A,descA,Rr,Rc,Sr,Sc,descRS) &
  bind (C,name='SDP_C_float_compress_ARS')
    ! Compression with an explicit matrix and user-given samples and random vectors
    import :: StrumpackDensePackage_F90_float
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_float) :: SDP_F90
    type(C_PTR), value :: A
    type(C_PTR), value :: descA
    type(C_PTR), value :: Rr
    type(C_PTR), value :: Rc
    type(C_PTR), value :: Sr
    type(C_PTR), value :: Sc
    type(C_PTR), value :: descRS
  end subroutine SDP_F90_float_compress_ARS

  subroutine SDP_F90_float_compress_mtxfree(SDP_F90,Rr,Rc,Sr,Sc,descRS,elements) &
  bind (C,name='SDP_C_float_compress_mtxfree')
    ! Matrix-free compression
    import :: StrumpackDensePackage_F90_float
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_float) :: SDP_F90
    type(C_PTR), value :: Rr
    type(C_PTR), value :: Rc
    type(C_PTR), value :: Sr
    type(C_PTR), value :: Sc
    type(C_PTR), value :: descRS
    type(C_PTR), value :: elements
  end subroutine SDP_F90_float_compress_mtxfree

  real(C_FLOAT) function SDP_F90_float_check_compression(SDP_F90,A,descA) &
  bind (C,name='SDP_C_float_check_compression')
    ! Compression quality check
    import :: StrumpackDensePackage_F90_float
    import :: C_FLOAT
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_float) :: SDP_F90
    type(C_PTR), value :: A
    type(C_PTR), value :: descA
  end function SDP_F90_float_check_compression

  subroutine SDP_F90_float_factor(SDP_F90,A,descA) &
  bind (C,name='SDP_C_float_factor')
    ! Factorization
    import :: StrumpackDensePackage_F90_float
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_float) :: SDP_F90
    type(C_PTR), value :: A
    type(C_PTR), value :: descA
  end subroutine SDP_F90_float_factor

  subroutine SDP_F90_float_partially_factor(SDP_F90,A,descA,nfact) &
  bind (C,name='SDP_C_float_partially_factor')
    ! Partial factorization
    import :: StrumpackDensePackage_F90_float
    import :: C_PTR
    import :: C_INT
    implicit none
    type(StrumpackDensePackage_F90_float) :: SDP_F90
    type(C_PTR), value    :: A
    type(C_PTR), value    :: descA
    integer(C_INT), value :: nfact
  end subroutine SDP_F90_float_partially_factor

  subroutine SDP_F90_float_compute_schur(SDP_F90) &
  bind (C,name='SDP_C_float_compute_schur')
    ! Schur complement computation
    import :: StrumpackDensePackage_F90_float
    implicit none
    type(StrumpackDensePackage_F90_float) :: SDP_F90
  end subroutine SDP_F90_float_compute_schur

  subroutine SDP_F90_float_reduce_RHS(SDP_F90,X,descX,B,descB) &
  bind (C,name='SDP_C_float_reduce_RHS')
    ! RHS reduction/condensation
    import :: StrumpackDensePackage_F90_float
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_float) :: SDP_F90
    type(C_PTR), value :: X
    type(C_PTR), value :: descX
    type(C_PTR), value :: B
    type(C_PTR), value :: descB
  end subroutine SDP_F90_float_reduce_RHS

  subroutine SDP_F90_float_schur_product(SDP_F90,Trans,alpha,X,descX,beta,B,descB) &
  bind (C,name='SDP_C_float_schur_product')
    ! Schur complement product
    import :: StrumpackDensePackage_F90_float
    import :: C_CHAR
    import :: C_FLOAT
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_float) :: SDP_F90
    character(C_CHAR), value :: Trans
    real(C_FLOAT), value    :: alpha
    type(C_PTR), value       :: X
    type(C_PTR), value       :: descX
    real(C_FLOAT), value    :: beta
    type(C_PTR), value       :: B
    type(C_PTR), value       :: descB
  end subroutine SDP_F90_float_schur_product

  subroutine SDP_F90_float_expand_solution(SDP_F90,X,descX,B,descB) &
  bind (C,name='SDP_C_float_expand_solution')
    ! Solution expansion
    import :: StrumpackDensePackage_F90_float
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_float) :: SDP_F90
    type(C_PTR), value :: X
    type(C_PTR), value :: descX
    type(C_PTR), value :: B
    type(C_PTR), value :: descB
  end subroutine SDP_F90_float_expand_solution

  subroutine SDP_F90_float_solve(SDP_F90,X,descX,B,descB) &
  bind (C,name='SDP_C_float_solve')
    ! Triangular solve
    import :: StrumpackDensePackage_F90_float
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_float) :: SDP_F90
    type(C_PTR), value :: X
    type(C_PTR), value :: descX
    type(C_PTR), value :: B
    type(C_PTR), value :: descB
  end subroutine SDP_F90_float_solve

  real(C_FLOAT) function SDP_F90_float_check_solution(SDP_F90,A,descA,X,descX,B,descB) &
  bind (C,name='SDP_C_float_check_solution')
    ! Solution accuracy check
    import :: StrumpackDensePackage_F90_float
    import :: C_FLOAT
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_float) :: SDP_F90
    type(C_PTR), value :: A
    type(C_PTR), value :: descA
    type(C_PTR), value :: X
    type(C_PTR), value :: descX
    type(C_PTR), value :: B
    type(C_PTR), value :: descB
  end function SDP_F90_float_check_solution

  real(C_FLOAT) function SDP_F90_float_refine(SDP_F90,A,descA,X,descX,B,descB) &
  bind (C,name='SDP_C_float_refine')
    ! Iterative refinement
    import :: StrumpackDensePackage_F90_float
    import :: C_FLOAT
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_float) :: SDP_F90
    type(C_PTR), value :: A
    type(C_PTR), value :: descA
    type(C_PTR), value :: X
    type(C_PTR), value :: descX
    type(C_PTR), value :: B
    type(C_PTR), value :: descB
  end function SDP_F90_float_refine

  subroutine SDP_F90_float_product(SDP_F90,Trans,alpha,A,descA,B,descB,beta,C,descC) &
  bind (C,name='SDP_C_float_product')
    ! Matrix-vector product
    import :: StrumpackDensePackage_F90_float
    import :: C_CHAR
    import :: C_FLOAT
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_float) :: SDP_F90
    character(C_CHAR), value :: Trans
    real(C_FLOAT), value    :: alpha
    type(C_PTR), value       :: A
    type(C_PTR), value       :: descA
    type(C_PTR), value       :: B
    type(C_PTR), value       :: descB
    real(C_FLOAT), value    :: beta
    type(C_PTR), value       :: C
    type(C_PTR), value       :: descC
  end subroutine SDP_F90_float_product

  subroutine SDP_F90_float_extract(SDP_F90,A,descA,B,descB,I,nI,J,nJ) &
  bind (C,name='SDP_C_float_extract')
    ! Element extraction
    import :: StrumpackDensePackage_F90_float
    import :: C_PTR
    import :: C_INT
    implicit none
    type(StrumpackDensePackage_F90_float) :: SDP_F90
    type(C_PTR), value    :: A
    type(C_PTR), value    :: descA
    type(C_PTR), value    :: B
    type(C_PTR), value    :: descB
    type(C_PTR), value    :: I
    integer(C_INT), value :: nI
    type(C_PTR), value    :: J
    integer(C_INT), value :: nJ
  end subroutine SDP_F90_float_extract

  subroutine SDP_F90_float_extract_schur(SDP_F90,B,descB,I,nI,J,nJ) &
  bind (C,name='SDP_C_float_extract_schur')
    ! Schur complement element extraction
    import :: StrumpackDensePackage_F90_float
    import :: C_PTR
    import :: C_INT
    implicit none
    type(StrumpackDensePackage_F90_float) :: SDP_F90
    type(C_PTR), value    :: B
    type(C_PTR), value    :: descB
    type(C_PTR), value    :: I
    integer(C_INT), value :: nI
    type(C_PTR), value    :: J
    integer(C_INT), value :: nJ
  end subroutine SDP_F90_float_extract_schur

  subroutine SDP_F90_float_print_statistics(SDP_F90) &
  bind (C,name='SDP_C_float_print_statistics')
    ! Statistics
    import :: StrumpackDensePackage_F90_float
    type(StrumpackDensePackage_F90_float) :: SDP_F90
  end subroutine SDP_F90_float_print_statistics

  !----------------------------------------------------------------------------
  ! Complex double precision

  subroutine SDP_F90_dcomplex_init(SDP_F90,i) &
  bind (C,name='SDP_C_dcomplex_init_F90')
    ! Initialization
    ! Here we don't call the C "init function" because MPI communicators do not
    ! have the same type in C (MPI_Comm) and Fortran (integer).
    import :: StrumpackDensePackage_F90_dcomplex
    import :: C_INT
    implicit none
    type(StrumpackDensePackage_F90_dcomplex) :: SDP_F90
    integer(C_INT), value :: i
  end subroutine SDP_F90_dcomplex_init

  subroutine SDP_F90_dcomplex_destroy(SDP_F90) &
  bind (C,name='SDP_C_dcomplex_destroy')
    ! Cleanup
    import :: StrumpackDensePackage_F90_dcomplex
    implicit none
    type(StrumpackDensePackage_F90_dcomplex) :: SDP_F90
  end subroutine SDP_F90_dcomplex_destroy

  subroutine SDP_F90_dcomplex_compress_A(SDP_F90,A,descA) &
  bind (C,name='SDP_C_dcomplex_compress_A')
    ! Compression with an explicit matrix
    import :: StrumpackDensePackage_F90_dcomplex
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_dcomplex) :: SDP_F90
    type(C_PTR), value :: A
    type(C_PTR), value :: descA
  end subroutine SDP_F90_dcomplex_compress_A

  subroutine SDP_F90_dcomplex_compress_ARS(SDP_F90,A,descA,Rr,Rc,Sr,Sc,descRS) &
  bind (C,name='SDP_C_dcomplex_compress_ARS')
    ! Compression with an explicit matrix and user-given samples and random vectors
    import :: StrumpackDensePackage_F90_dcomplex
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_dcomplex) :: SDP_F90
    type(C_PTR), value :: A
    type(C_PTR), value :: descA
    type(C_PTR), value :: Rr
    type(C_PTR), value :: Rc
    type(C_PTR), value :: Sr
    type(C_PTR), value :: Sc
    type(C_PTR), value :: descRS
  end subroutine SDP_F90_dcomplex_compress_ARS

  subroutine SDP_F90_dcomplex_compress_mtxfree(SDP_F90,Rr,Rc,Sr,Sc,descRS,elements) &
  bind (C,name='SDP_C_dcomplex_compress_mtxfree')
    ! Matrix-free compression
    import :: StrumpackDensePackage_F90_dcomplex
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_dcomplex) :: SDP_F90
    type(C_PTR), value :: Rr
    type(C_PTR), value :: Rc
    type(C_PTR), value :: Sr
    type(C_PTR), value :: Sc
    type(C_PTR), value :: descRS
    type(C_PTR), value :: elements
  end subroutine SDP_F90_dcomplex_compress_mtxfree

  real(C_DOUBLE) function SDP_F90_dcomplex_check_compression(SDP_F90,A,descA) &
  bind (C,name='SDP_C_dcomplex_check_compression')
    ! Compression quality check
    import :: StrumpackDensePackage_F90_dcomplex
    import :: C_DOUBLE
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_dcomplex) :: SDP_F90
    type(C_PTR), value :: A
    type(C_PTR), value :: descA
  end function SDP_F90_dcomplex_check_compression

  subroutine SDP_F90_dcomplex_factor(SDP_F90,A,descA) &
  bind (C,name='SDP_C_dcomplex_factor')
    ! Factorization
    import :: StrumpackDensePackage_F90_dcomplex
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_dcomplex) :: SDP_F90
    type(C_PTR), value :: A
    type(C_PTR), value :: descA
  end subroutine SDP_F90_dcomplex_factor

  subroutine SDP_F90_dcomplex_partially_factor(SDP_F90,A,descA,nfact) &
  bind (C,name='SDP_C_dcomplex_partially_factor')
    ! Partial factorization
    import :: StrumpackDensePackage_F90_dcomplex
    import :: C_PTR
    import :: C_INT
    implicit none
    type(StrumpackDensePackage_F90_dcomplex) :: SDP_F90
    type(C_PTR), value    :: A
    type(C_PTR), value    :: descA
    integer(C_INT), value :: nfact
  end subroutine SDP_F90_dcomplex_partially_factor

  subroutine SDP_F90_dcomplex_compute_schur(SDP_F90) &
  bind (C,name='SDP_C_dcomplex_compute_schur')
    ! Schur complement computation
    import :: StrumpackDensePackage_F90_dcomplex
    implicit none
    type(StrumpackDensePackage_F90_dcomplex) :: SDP_F90
  end subroutine SDP_F90_dcomplex_compute_schur

  subroutine SDP_F90_dcomplex_reduce_RHS(SDP_F90,X,descX,B,descB) &
  bind (C,name='SDP_C_dcomplex_reduce_RHS')
    ! RHS reduction/condensation
    import :: StrumpackDensePackage_F90_dcomplex
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_dcomplex) :: SDP_F90
    type(C_PTR), value :: X
    type(C_PTR), value :: descX
    type(C_PTR), value :: B
    type(C_PTR), value :: descB
  end subroutine SDP_F90_dcomplex_reduce_RHS

  subroutine SDP_F90_dcomplex_schur_product(SDP_F90,Trans,alpha,X,descX,beta,B,descB) &
  bind (C,name='SDP_C_dcomplex_schur_product')
    ! Schur complement product
    import :: StrumpackDensePackage_F90_dcomplex
    import :: C_CHAR
    import :: C_DOUBLE_COMPLEX
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_dcomplex) :: SDP_F90
    character(C_CHAR), value :: Trans
    real(C_DOUBLE_COMPLEX), value :: alpha
    type(C_PTR), value            :: X
    type(C_PTR), value            :: descX
    real(C_DOUBLE_COMPLEX), value :: beta
    type(C_PTR), value            :: B
    type(C_PTR), value            :: descB
  end subroutine SDP_F90_dcomplex_schur_product

  subroutine SDP_F90_dcomplex_expand_solution(SDP_F90,X,descX,B,descB) &
  bind (C,name='SDP_C_dcomplex_expand_solution')
    ! Solution expansion
    import :: StrumpackDensePackage_F90_dcomplex
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_dcomplex) :: SDP_F90
    type(C_PTR), value :: X
    type(C_PTR), value :: descX
    type(C_PTR), value :: B
    type(C_PTR), value :: descB
  end subroutine SDP_F90_dcomplex_expand_solution

  subroutine SDP_F90_dcomplex_solve(SDP_F90,X,descX,B,descB) &
  bind (C,name='SDP_C_dcomplex_solve')
    ! Triangular solve
    import :: StrumpackDensePackage_F90_dcomplex
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_dcomplex) :: SDP_F90
    type(C_PTR), value :: X
    type(C_PTR), value :: descX
    type(C_PTR), value :: B
    type(C_PTR), value :: descB
  end subroutine SDP_F90_dcomplex_solve

  real(C_DOUBLE) function SDP_F90_dcomplex_check_solution(SDP_F90,A,descA,X,descX,B,descB) &
  bind (C,name='SDP_C_dcomplex_check_solution')
    ! Solution accuracy check
    import :: StrumpackDensePackage_F90_dcomplex
    import :: C_DOUBLE
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_dcomplex) :: SDP_F90
    type(C_PTR), value :: A
    type(C_PTR), value :: descA
    type(C_PTR), value :: X
    type(C_PTR), value :: descX
    type(C_PTR), value :: B
    type(C_PTR), value :: descB
  end function SDP_F90_dcomplex_check_solution

  real(C_DOUBLE) function SDP_F90_dcomplex_refine(SDP_F90,A,descA,X,descX,B,descB) &
  bind (C,name='SDP_C_dcomplex_refine')
    ! Iterative refinement
    import :: StrumpackDensePackage_F90_dcomplex
    import :: C_DOUBLE
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_dcomplex) :: SDP_F90
    type(C_PTR), value :: A
    type(C_PTR), value :: descA
    type(C_PTR), value :: X
    type(C_PTR), value :: descX
    type(C_PTR), value :: B
    type(C_PTR), value :: descB
  end function SDP_F90_dcomplex_refine

  subroutine SDP_F90_dcomplex_product(SDP_F90,Trans,alpha,A,descA,B,descB,beta,C,descC) &
  bind (C,name='SDP_C_dcomplex_product')
    ! Matrix-vector product
    import :: StrumpackDensePackage_F90_dcomplex
    import :: C_CHAR
    import :: C_DOUBLE_COMPLEX
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_dcomplex) :: SDP_F90
    character(C_CHAR), value :: Trans
    real(C_DOUBLE_COMPLEX), value :: alpha
    type(C_PTR), value            :: A
    type(C_PTR), value            :: descA
    type(C_PTR), value            :: B
    type(C_PTR), value            :: descB
    real(C_DOUBLE_COMPLEX), value :: beta
    type(C_PTR), value            :: C
    type(C_PTR), value            :: descC
  end subroutine SDP_F90_dcomplex_product

  subroutine SDP_F90_dcomplex_extract(SDP_F90,A,descA,B,descB,I,nI,J,nJ) &
  bind (C,name='SDP_C_dcomplex_extract')
    ! Element extraction
    import :: StrumpackDensePackage_F90_dcomplex
    import :: C_PTR
    import :: C_INT
    implicit none
    type(StrumpackDensePackage_F90_dcomplex) :: SDP_F90
    type(C_PTR), value    :: A
    type(C_PTR), value    :: descA
    type(C_PTR), value    :: B
    type(C_PTR), value    :: descB
    type(C_PTR), value    :: I
    integer(C_INT), value :: nI
    type(C_PTR), value    :: J
    integer(C_INT), value :: nJ
  end subroutine SDP_F90_dcomplex_extract

  subroutine SDP_F90_dcomplex_extract_schur(SDP_F90,B,descB,I,nI,J,nJ) &
  bind (C,name='SDP_C_dcomplex_extract_schur')
    ! Schur complement element extraction
    import :: StrumpackDensePackage_F90_dcomplex
    import :: C_PTR
    import :: C_INT
    implicit none
    type(StrumpackDensePackage_F90_dcomplex) :: SDP_F90
    type(C_PTR), value    :: B
    type(C_PTR), value    :: descB
    type(C_PTR), value    :: I
    integer(C_INT), value :: nI
    type(C_PTR), value    :: J
    integer(C_INT), value :: nJ
  end subroutine SDP_F90_dcomplex_extract_schur

  subroutine SDP_F90_dcomplex_print_statistics(SDP_F90) &
  bind (C,name='SDP_C_dcomplex_print_statistics')
    ! Statistics
    import :: StrumpackDensePackage_F90_dcomplex
    type(StrumpackDensePackage_F90_dcomplex) :: SDP_F90
  end subroutine SDP_F90_dcomplex_print_statistics

  !----------------------------------------------------------------------------
  ! Complex single precision

  subroutine SDP_F90_scomplex_init(SDP_F90,i) &
  bind (C,name='SDP_C_scomplex_init_F90')
    ! Initialization
    ! Here we don't call the C "init function" because MPI communicators do not
    ! have the same type in C (MPI_Comm) and Fortran (integer).
    import :: StrumpackDensePackage_F90_scomplex
    import :: C_INT
    implicit none
    type(StrumpackDensePackage_F90_scomplex) :: SDP_F90
    integer(C_INT), value :: i
  end subroutine SDP_F90_scomplex_init

  subroutine SDP_F90_scomplex_destroy(SDP_F90) &
  bind (C,name='SDP_C_scomplex_destroy')
    ! Cleanup
    import :: StrumpackDensePackage_F90_scomplex
    implicit none
    type(StrumpackDensePackage_F90_scomplex) :: SDP_F90
  end subroutine SDP_F90_scomplex_destroy

  subroutine SDP_F90_scomplex_compress_A(SDP_F90,A,descA) &
  bind (C,name='SDP_C_scomplex_compress_A')
    ! Compression with an explicit matrix
    import :: StrumpackDensePackage_F90_scomplex
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_scomplex) :: SDP_F90
    type(C_PTR), value :: A
    type(C_PTR), value :: descA
  end subroutine SDP_F90_scomplex_compress_A

  subroutine SDP_F90_scomplex_compress_ARS(SDP_F90,A,descA,Rr,Rc,Sr,Sc,descRS) &
  bind (C,name='SDP_C_scomplex_compress_ARS')
    ! Compression with an explicit matrix and user-given samples and random vectors
    import :: StrumpackDensePackage_F90_scomplex
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_scomplex) :: SDP_F90
    type(C_PTR), value :: A
    type(C_PTR), value :: descA
    type(C_PTR), value :: Rr
    type(C_PTR), value :: Rc
    type(C_PTR), value :: Sr
    type(C_PTR), value :: Sc
    type(C_PTR), value :: descRS
  end subroutine SDP_F90_scomplex_compress_ARS

  subroutine SDP_F90_scomplex_compress_mtxfree(SDP_F90,Rr,Rc,Sr,Sc,descRS,elements) &
  bind (C,name='SDP_C_scomplex_compress_mtxfree')
    ! Matrix-free compression
    import :: StrumpackDensePackage_F90_scomplex
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_scomplex) :: SDP_F90
    type(C_PTR), value :: Rr
    type(C_PTR), value :: Rc
    type(C_PTR), value :: Sr
    type(C_PTR), value :: Sc
    type(C_PTR), value :: descRS
    type(C_PTR), value :: elements
  end subroutine SDP_F90_scomplex_compress_mtxfree

  real(C_FLOAT) function SDP_F90_scomplex_check_compression(SDP_F90,A,descA) &
  bind (C,name='SDP_C_scomplex_check_compression')
    ! Compression quality check
    import :: StrumpackDensePackage_F90_scomplex
    import :: C_FLOAT
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_scomplex) :: SDP_F90
    type(C_PTR), value :: A
    type(C_PTR), value :: descA
  end function SDP_F90_scomplex_check_compression

  subroutine SDP_F90_scomplex_factor(SDP_F90,A,descA) &
  bind (C,name='SDP_C_scomplex_factor')
    ! Factorization
    import :: StrumpackDensePackage_F90_scomplex
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_scomplex) :: SDP_F90
    type(C_PTR), value :: A
    type(C_PTR), value :: descA
  end subroutine SDP_F90_scomplex_factor

  subroutine SDP_F90_scomplex_partially_factor(SDP_F90,A,descA,nfact) &
  bind (C,name='SDP_C_scomplex_partially_factor')
    ! Partial factorization
    import :: StrumpackDensePackage_F90_scomplex
    import :: C_PTR
    import :: C_INT
    implicit none
    type(StrumpackDensePackage_F90_scomplex) :: SDP_F90
    type(C_PTR), value    :: A
    type(C_PTR), value    :: descA
    integer(C_INT), value :: nfact
  end subroutine SDP_F90_scomplex_partially_factor

  subroutine SDP_F90_scomplex_compute_schur(SDP_F90) &
  bind (C,name='SDP_C_scomplex_compute_schur')
    ! Schur complement computation
    import :: StrumpackDensePackage_F90_scomplex
    implicit none
    type(StrumpackDensePackage_F90_scomplex) :: SDP_F90
  end subroutine SDP_F90_scomplex_compute_schur

  subroutine SDP_F90_scomplex_reduce_RHS(SDP_F90,X,descX,B,descB) &
  bind (C,name='SDP_C_scomplex_reduce_RHS')
    ! RHS reduction/condensation
    import :: StrumpackDensePackage_F90_scomplex
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_scomplex) :: SDP_F90
    type(C_PTR), value :: X
    type(C_PTR), value :: descX
    type(C_PTR), value :: B
    type(C_PTR), value :: descB
  end subroutine SDP_F90_scomplex_reduce_RHS

  subroutine SDP_F90_scomplex_schur_product(SDP_F90,Trans,alpha,X,descX,beta,B,descB) &
  bind (C,name='SDP_C_scomplex_schur_product')
    ! Schur complement product
    import :: StrumpackDensePackage_F90_scomplex
    import :: C_CHAR
    import :: C_FLOAT_COMPLEX
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_scomplex) :: SDP_F90
    character(C_CHAR), value :: Trans
    real(C_FLOAT_COMPLEX), value :: alpha
    type(C_PTR), value            :: X
    type(C_PTR), value            :: descX
    real(C_FLOAT_COMPLEX), value :: beta
    type(C_PTR), value            :: B
    type(C_PTR), value            :: descB
  end subroutine SDP_F90_scomplex_schur_product

  subroutine SDP_F90_scomplex_expand_solution(SDP_F90,X,descX,B,descB) &
  bind (C,name='SDP_C_scomplex_expand_solution')
    ! Solution expansion
    import :: StrumpackDensePackage_F90_scomplex
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_scomplex) :: SDP_F90
    type(C_PTR), value :: X
    type(C_PTR), value :: descX
    type(C_PTR), value :: B
    type(C_PTR), value :: descB
  end subroutine SDP_F90_scomplex_expand_solution

  subroutine SDP_F90_scomplex_solve(SDP_F90,X,descX,B,descB) &
  bind (C,name='SDP_C_scomplex_solve')
    ! Triangular solve
    import :: StrumpackDensePackage_F90_scomplex
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_scomplex) :: SDP_F90
    type(C_PTR), value :: X
    type(C_PTR), value :: descX
    type(C_PTR), value :: B
    type(C_PTR), value :: descB
  end subroutine SDP_F90_scomplex_solve

  real(C_FLOAT) function SDP_F90_scomplex_check_solution(SDP_F90,A,descA,X,descX,B,descB) &
  bind (C,name='SDP_C_scomplex_check_solution')
    ! Solution accuracy check
    import :: StrumpackDensePackage_F90_scomplex
    import :: C_FLOAT
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_scomplex) :: SDP_F90
    type(C_PTR), value :: A
    type(C_PTR), value :: descA
    type(C_PTR), value :: X
    type(C_PTR), value :: descX
    type(C_PTR), value :: B
    type(C_PTR), value :: descB
  end function SDP_F90_scomplex_check_solution

  real(C_FLOAT) function SDP_F90_scomplex_refine(SDP_F90,A,descA,X,descX,B,descB) &
  bind (C,name='SDP_C_scomplex_refine')
    ! Iterative refinement
    import :: StrumpackDensePackage_F90_scomplex
    import :: C_FLOAT
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_scomplex) :: SDP_F90
    type(C_PTR), value :: A
    type(C_PTR), value :: descA
    type(C_PTR), value :: X
    type(C_PTR), value :: descX
    type(C_PTR), value :: B
    type(C_PTR), value :: descB
  end function SDP_F90_scomplex_refine

  subroutine SDP_F90_scomplex_product(SDP_F90,Trans,alpha,A,descA,B,descB,beta,C,descC) &
  bind (C,name='SDP_C_scomplex_product')
    ! Matrix-vector product
    import :: StrumpackDensePackage_F90_scomplex
    import :: C_CHAR
    import :: C_FLOAT_COMPLEX
    import :: C_PTR
    implicit none
    type(StrumpackDensePackage_F90_scomplex) :: SDP_F90
    character(C_CHAR), value :: Trans
    real(C_FLOAT_COMPLEX), value :: alpha
    type(C_PTR), value            :: A
    type(C_PTR), value            :: descA
    type(C_PTR), value            :: B
    type(C_PTR), value            :: descB
    real(C_FLOAT_COMPLEX), value :: beta
    type(C_PTR), value            :: C
    type(C_PTR), value            :: descC
  end subroutine SDP_F90_scomplex_product

  subroutine SDP_F90_scomplex_extract(SDP_F90,A,descA,B,descB,I,nI,J,nJ) &
  bind (C,name='SDP_C_scomplex_extract')
    ! Element extraction
    import :: StrumpackDensePackage_F90_scomplex
    import :: C_PTR
    import :: C_INT
    implicit none
    type(StrumpackDensePackage_F90_scomplex) :: SDP_F90
    type(C_PTR), value    :: A
    type(C_PTR), value    :: descA
    type(C_PTR), value    :: B
    type(C_PTR), value    :: descB
    type(C_PTR), value    :: I
    integer(C_INT), value :: nI
    type(C_PTR), value    :: J
    integer(C_INT), value :: nJ
  end subroutine SDP_F90_scomplex_extract

  subroutine SDP_F90_scomplex_extract_schur(SDP_F90,B,descB,I,nI,J,nJ) &
  bind (C,name='SDP_C_scomplex_extract_schur')
    ! Schur complement element extraction
    import :: StrumpackDensePackage_F90_scomplex
    import :: C_PTR
    import :: C_INT
    implicit none
    type(StrumpackDensePackage_F90_scomplex) :: SDP_F90
    type(C_PTR), value    :: B
    type(C_PTR), value    :: descB
    type(C_PTR), value    :: I
    integer(C_INT), value :: nI
    type(C_PTR), value    :: J
    integer(C_INT), value :: nJ
  end subroutine SDP_F90_scomplex_extract_schur

  subroutine SDP_F90_scomplex_print_statistics(SDP_F90) &
  bind (C,name='SDP_C_scomplex_print_statistics')
    ! Statistics
    import :: StrumpackDensePackage_F90_scomplex
    type(StrumpackDensePackage_F90_scomplex) :: SDP_F90
  end subroutine SDP_F90_scomplex_print_statistics

end interface

end module StrumpackDensePackage
