/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */

#include <stdlib.h>
#include <math.h>

/* This C++ file implements the functions of the C interface. */
#include "StrumpackDensePackage.hpp"

extern "C" {
  #include "StrumpackDensePackage.h"

/*****************************************************************************/
  /* Real double precision */
  void SDP_C_double_copy_fields(StrumpackDensePackage_C_double *SDP_C) {
    /* Copy all the fields from the C structure to the C++ class */
    StrumpackDensePackage<double,double>* ptr=static_cast<StrumpackDensePackage<double,double>*>(SDP_C->SDP);
    ptr->use_HSS         = SDP_C->use_HSS;
    ptr->tol_HSS         = SDP_C->tol_HSS;
    ptr->min_rand_HSS    = SDP_C->min_rand_HSS;
    ptr->max_rand_HSS    = SDP_C->max_rand_HSS;
    ptr->lim_rand_HSS    = SDP_C->lim_rand_HSS;
    ptr->inc_rand_HSS    = SDP_C->inc_rand_HSS;
    ptr->block_HSS       = SDP_C->block_HSS;
    ptr->levels_HSS      = SDP_C->levels_HSS;
    ptr->split_HSS       = SDP_C->split_HSS;
    ptr->tol_IR          = SDP_C->tol_IR;
    ptr->steps_IR        = SDP_C->steps_IR;
    ptr->fast_IR         = SDP_C->fast_IR;
    ptr->verbose         = SDP_C->verbose;
    ptr->access_is_global= SDP_C->access_is_global;
    ptr->sched_strat     = SDP_C->sched_strat;
    ptr->obj             = SDP_C->obj;
  }

  void SDP_C_double_init(StrumpackDensePackage_C_double *SDP_C, MPI_Comm comm) {
    /* Initialization: calls the C++ constructor */
    SDP_C->SDP=static_cast<void*>(new StrumpackDensePackage<double,double>(comm));
    /* Copy default options to C structure */
    StrumpackDensePackage<double,double>* ptr=static_cast<StrumpackDensePackage<double,double>*>(SDP_C->SDP);
    SDP_C->use_HSS         = ptr->use_HSS;
    SDP_C->tol_HSS         = ptr->tol_HSS;
    SDP_C->min_rand_HSS    = ptr->min_rand_HSS;
    SDP_C->max_rand_HSS    = ptr->max_rand_HSS;
    SDP_C->lim_rand_HSS    = ptr->lim_rand_HSS;
    SDP_C->inc_rand_HSS    = ptr->inc_rand_HSS;
    SDP_C->block_HSS       = ptr->block_HSS;
    SDP_C->levels_HSS      = ptr->levels_HSS;
    SDP_C->split_HSS       = ptr->split_HSS;
    SDP_C->tol_IR          = ptr->tol_IR;
    SDP_C->steps_IR        = ptr->steps_IR;
    SDP_C->fast_IR         = ptr->fast_IR;
    SDP_C->verbose         = ptr->verbose;
    SDP_C->access_is_global= ptr->access_is_global;
    SDP_C->sched_strat     = ptr->sched_strat;
    SDP_C->obj             = ptr->obj;
  }

  void SDP_C_double_init_F90(StrumpackDensePackage_C_double *SDP_C, int fcomm) {
    /* Initialization: calls the C++ constructor */
    SDP_C->SDP=static_cast<void*>(new StrumpackDensePackage<double,double>(MPI_Comm_f2c(fcomm)));
    /* Copy default options to C structure */
    StrumpackDensePackage<double,double>* ptr=static_cast<StrumpackDensePackage<double,double>*>(SDP_C->SDP);
    SDP_C->use_HSS         = ptr->use_HSS;
    SDP_C->tol_HSS         = ptr->tol_HSS;
    SDP_C->min_rand_HSS    = ptr->min_rand_HSS;
    SDP_C->max_rand_HSS    = ptr->max_rand_HSS;
    SDP_C->lim_rand_HSS    = ptr->lim_rand_HSS;
    SDP_C->inc_rand_HSS    = ptr->inc_rand_HSS;
    SDP_C->block_HSS       = ptr->block_HSS;
    SDP_C->levels_HSS      = ptr->levels_HSS;
    SDP_C->split_HSS       = ptr->split_HSS;
    SDP_C->tol_IR          = ptr->tol_IR;
    SDP_C->steps_IR        = ptr->steps_IR;
    SDP_C->fast_IR         = ptr->fast_IR;
    SDP_C->verbose         = ptr->verbose;
    SDP_C->access_is_global= ptr->access_is_global;
    SDP_C->sched_strat     = ptr->sched_strat;
    SDP_C->obj             = ptr->obj;
  }

  void SDP_C_double_destroy(StrumpackDensePackage_C_double *SDP_C) {
    /* Cleanup: calls to C++ destructor */
    StrumpackDensePackage<double,double>* ptr=static_cast<StrumpackDensePackage<double,double>*>(SDP_C->SDP);
    delete ptr;
  }

  void SDP_C_double_compress_A(StrumpackDensePackage_C_double *SDP_C, double *A, int *descA) {
    /* Compression with an explicit matrix */
    StrumpackDensePackage<double,double>* ptr=static_cast<StrumpackDensePackage<double,double>*>(SDP_C->SDP);
    SDP_C_double_copy_fields(SDP_C);
    ptr->compress(A,descA);
  }

  void SDP_C_double_compress_ARS(StrumpackDensePackage_C_double *SDP_C, double* A, int *descA, double *Rr, double *Rc, double* Sr, double* Sc, int *descRS) {
    /* Compression with an explicit matrix and user-given samples and random vectors */
    StrumpackDensePackage<double,double>* ptr=static_cast<StrumpackDensePackage<double,double>*>(SDP_C->SDP);
    SDP_C_double_copy_fields(SDP_C);
    ptr->compress(A,descA,Rr,Rc,Sr,Sc,descRS);
  }

  void SDP_C_double_compress_mtxfree(StrumpackDensePackage_C_double *SDP_C, double *Rr, double*Rc, double*Sr, double* Sc, int *descRS, void (*elements)(void *, int *, int *, double *, int *)) {
    /* Matrix-free compression */
    StrumpackDensePackage<double,double>* ptr=static_cast<StrumpackDensePackage<double,double>*>(SDP_C->SDP);
    SDP_C_double_copy_fields(SDP_C);
    ptr->compress(Rr,Rc,Sr,Sc,descRS,elements);
  }

  double SDP_C_double_check_compression(StrumpackDensePackage_C_double *SDP_C, double *A, int *descA) {
    /* Compression quality check */
    StrumpackDensePackage<double,double>* ptr=static_cast<StrumpackDensePackage<double,double>*>(SDP_C->SDP);
    SDP_C_double_copy_fields(SDP_C);
    return ptr->check_compression(A,descA);
  }

  void SDP_C_double_factor(StrumpackDensePackage_C_double *SDP_C, double *A, int *descA) {
    /* Factorization */
    StrumpackDensePackage<double,double>* ptr=static_cast<StrumpackDensePackage<double,double>*>(SDP_C->SDP);
    SDP_C_double_copy_fields(SDP_C);
    ptr->factor(A,descA);
  }

  void SDP_C_double_partially_factor(StrumpackDensePackage_C_double *SDP_C, double* A, int *descA, int nfact) {
    /* Partial factorization */
    StrumpackDensePackage<double,double>* ptr=static_cast<StrumpackDensePackage<double,double>*>(SDP_C->SDP);
    SDP_C_double_copy_fields(SDP_C);
    ptr->partially_factor(A,descA,nfact);
  }

  void SDP_C_double_compute_schur(StrumpackDensePackage_C_double *SDP_C){
    /* Schur complement computation */
    StrumpackDensePackage<double,double>* ptr=static_cast<StrumpackDensePackage<double,double>*>(SDP_C->SDP);
    SDP_C_double_copy_fields(SDP_C);
    ptr->compute_schur();
  }

  void SDP_C_double_reduce_RHS(StrumpackDensePackage_C_double *SDP_C, double *X, int *descX, double* B, int *descB){
    /* RHS reduction/condensation */
    StrumpackDensePackage<double,double>* ptr=static_cast<StrumpackDensePackage<double,double>*>(SDP_C->SDP);
    SDP_C_double_copy_fields(SDP_C);
    ptr->reduce_RHS(X,descX,B,descB);
  }

  void SDP_C_double_schur_product(StrumpackDensePackage_C_double *SDP_C, char Trans, double alpha, double *X, int *descX, double beta, double *B, int *descB){
    /* Schur complement product */
    StrumpackDensePackage<double,double>* ptr=static_cast<StrumpackDensePackage<double,double>*>(SDP_C->SDP);
    SDP_C_double_copy_fields(SDP_C);
    ptr->schur_product(Trans,alpha,X,descX,beta,B,descB);
  }

  void SDP_C_double_expand_solution(StrumpackDensePackage_C_double *SDP_C, double *X, int *descX, double *B, int *descB){
    /* Solution expansion */
    StrumpackDensePackage<double,double>* ptr=static_cast<StrumpackDensePackage<double,double>*>(SDP_C->SDP);
    SDP_C_double_copy_fields(SDP_C);
    ptr->expand_solution(X,descX,B,descB);
  }

  void SDP_C_double_solve(StrumpackDensePackage_C_double *SDP_C, double *X, int *descX, double *B, int *descB){
    /* Triangular solve */
    StrumpackDensePackage<double,double>* ptr=static_cast<StrumpackDensePackage<double,double>*>(SDP_C->SDP);
    SDP_C_double_copy_fields(SDP_C);
    ptr->solve(X,descX,B,descB);
  }

  double SDP_C_double_check_solution(StrumpackDensePackage_C_double *SDP_C, double *A, int *descA, double *X, int *descX, double *B, int *descB){
    /* Solution accuracy check */
    StrumpackDensePackage<double,double>* ptr=static_cast<StrumpackDensePackage<double,double>*>(SDP_C->SDP);
    SDP_C_double_copy_fields(SDP_C);
    return ptr->check_solution(A,descA,X,descX,B,descB);
  }

  double SDP_C_double_refine(StrumpackDensePackage_C_double *SDP_C, double *A, int *descA, double *X, int *descX, double *B, int *descB){
    /* Iterative refinement */
    StrumpackDensePackage<double,double>* ptr=static_cast<StrumpackDensePackage<double,double>*>(SDP_C->SDP);
    SDP_C_double_copy_fields(SDP_C);
    return ptr->refine(A,descA,X,descX,B,descB);
  }

  void SDP_C_double_product(StrumpackDensePackage_C_double *SDP_C, char Trans, double alpha, double *A, int *descA, double *B, int *descB, double beta, double *C, int *descC){
    /* Matrix-vector product */
    StrumpackDensePackage<double,double>* ptr=static_cast<StrumpackDensePackage<double,double>*>(SDP_C->SDP);
    SDP_C_double_copy_fields(SDP_C);
    ptr->product(Trans,alpha,A,descA,B,descB,beta,C,descC);
  }

  void SDP_C_double_extract(StrumpackDensePackage_C_double *SDP_C, double *A, int *descA, double *B, int *descB, int *I, int nI, int *J, int nJ){
    /* Element extraction */
    StrumpackDensePackage<double,double>* ptr=static_cast<StrumpackDensePackage<double,double>*>(SDP_C->SDP);
    SDP_C_double_copy_fields(SDP_C);
    ptr->extract(A,descA,B,descB,I,nI,J,nJ);
  }

  void SDP_C_double_extract_schur(StrumpackDensePackage_C_double *SDP_C, double *B, int *descB, int *I, int nI, int *J, int nJ){
    /* Schur complement element extraction */
    StrumpackDensePackage<double,double>* ptr=static_cast<StrumpackDensePackage<double,double>*>(SDP_C->SDP);
    SDP_C_double_copy_fields(SDP_C);
    ptr->extract_schur(B,descB,I,nI,J,nJ);
  }

  void SDP_C_double_print_statistics(StrumpackDensePackage_C_double *SDP_C){
    /* Statistics */
    StrumpackDensePackage<double,double>* ptr=static_cast<StrumpackDensePackage<double,double>*>(SDP_C->SDP);
    SDP_C_double_copy_fields(SDP_C);
    ptr->print_statistics();
  }

/*****************************************************************************/
  /* Real single precision */
  void SDP_C_float_copy_fields(StrumpackDensePackage_C_float *SDP_C) {
    /* Copy all the fields from the C structure to the C++ class */
    StrumpackDensePackage<float,float>* ptr=static_cast<StrumpackDensePackage<float,float>*>(SDP_C->SDP);
    ptr->use_HSS         = SDP_C->use_HSS;
    ptr->tol_HSS         = SDP_C->tol_HSS;
    ptr->min_rand_HSS    = SDP_C->min_rand_HSS;
    ptr->max_rand_HSS    = SDP_C->max_rand_HSS;
    ptr->lim_rand_HSS    = SDP_C->lim_rand_HSS;
    ptr->inc_rand_HSS    = SDP_C->inc_rand_HSS;
    ptr->block_HSS       = SDP_C->block_HSS;
    ptr->levels_HSS      = SDP_C->levels_HSS;
    ptr->split_HSS       = SDP_C->split_HSS;
    ptr->tol_IR          = SDP_C->tol_IR;
    ptr->steps_IR        = SDP_C->steps_IR;
    ptr->fast_IR         = SDP_C->fast_IR;
    ptr->verbose         = SDP_C->verbose;
    ptr->access_is_global= SDP_C->access_is_global;
    ptr->sched_strat     = SDP_C->sched_strat;
    ptr->obj             = SDP_C->obj;
  }

  void SDP_C_float_init(StrumpackDensePackage_C_float *SDP_C, MPI_Comm comm) {
    /* Initialization: calls the C++ constructor */
    SDP_C->SDP=static_cast<void*>(new StrumpackDensePackage<float,float>(comm));
    /* Copy default options to C structure */
    StrumpackDensePackage<float,float>* ptr=static_cast<StrumpackDensePackage<float,float>*>(SDP_C->SDP);
    SDP_C->use_HSS         = ptr->use_HSS;
    SDP_C->tol_HSS         = ptr->tol_HSS;
    SDP_C->min_rand_HSS    = ptr->min_rand_HSS;
    SDP_C->max_rand_HSS    = ptr->max_rand_HSS;
    SDP_C->lim_rand_HSS    = ptr->lim_rand_HSS;
    SDP_C->inc_rand_HSS    = ptr->inc_rand_HSS;
    SDP_C->block_HSS       = ptr->block_HSS;
    SDP_C->levels_HSS      = ptr->levels_HSS;
    SDP_C->split_HSS       = ptr->split_HSS;
    SDP_C->tol_IR          = ptr->tol_IR;
    SDP_C->steps_IR        = ptr->steps_IR;
    SDP_C->fast_IR         = ptr->fast_IR;
    SDP_C->verbose         = ptr->verbose;
    SDP_C->access_is_global= ptr->access_is_global;
    SDP_C->sched_strat     = ptr->sched_strat;
    SDP_C->obj             = ptr->obj;
  }

  void SDP_C_float_destroy(StrumpackDensePackage_C_float *SDP_C) {
    /* Cleanup: calls to C++ destructor */
    StrumpackDensePackage<float,float>* ptr=static_cast<StrumpackDensePackage<float,float>*>(SDP_C->SDP);
    delete ptr;
  }

  void SDP_C_float_compress_A(StrumpackDensePackage_C_float *SDP_C, float *A, int *descA) {
    /* Compression with an explicit matrix */
    StrumpackDensePackage<float,float>* ptr=static_cast<StrumpackDensePackage<float,float>*>(SDP_C->SDP);
    SDP_C_float_copy_fields(SDP_C);
    ptr->compress(A,descA);
  }

  void SDP_C_float_compress_ARS(StrumpackDensePackage_C_float *SDP_C, float* A, int *descA, float *Rr, float *Rc, float* Sr, float* Sc, int *descRS) {
    /* Compression with an explicit matrix and user-given samples and random vectors */
    StrumpackDensePackage<float,float>* ptr=static_cast<StrumpackDensePackage<float,float>*>(SDP_C->SDP);
    SDP_C_float_copy_fields(SDP_C);
    ptr->compress(A,descA,Rr,Rc,Sr,Sc,descRS);
  }

  void SDP_C_float_compress_mtxfree(StrumpackDensePackage_C_float *SDP_C, float *Rr, float*Rc, float*Sr, float* Sc, int *descRS, void (*elements)(void *, int *, int *, float *, int *)) {
    /* Matrix-free compression */
    StrumpackDensePackage<float,float>* ptr=static_cast<StrumpackDensePackage<float,float>*>(SDP_C->SDP);
    SDP_C_float_copy_fields(SDP_C);
    ptr->compress(Rr,Rc,Sr,Sc,descRS,elements);
  }

  float SDP_C_float_check_compression(StrumpackDensePackage_C_float *SDP_C, float *A, int *descA) {
    /* Compression quality check */
    StrumpackDensePackage<float,float>* ptr=static_cast<StrumpackDensePackage<float,float>*>(SDP_C->SDP);
    SDP_C_float_copy_fields(SDP_C);
    return ptr->check_compression(A,descA);
  }

  void SDP_C_float_factor(StrumpackDensePackage_C_float *SDP_C, float *A, int *descA) {
    /* Factorization */
    StrumpackDensePackage<float,float>* ptr=static_cast<StrumpackDensePackage<float,float>*>(SDP_C->SDP);
    SDP_C_float_copy_fields(SDP_C);
    ptr->factor(A,descA);
  }

  void SDP_C_float_partially_factor(StrumpackDensePackage_C_float *SDP_C, float* A, int *descA, int nfact) {
    /* Partial factorization */
    StrumpackDensePackage<float,float>* ptr=static_cast<StrumpackDensePackage<float,float>*>(SDP_C->SDP);
    SDP_C_float_copy_fields(SDP_C);
    ptr->partially_factor(A,descA,nfact);
  }

  void SDP_C_float_compute_schur(StrumpackDensePackage_C_float *SDP_C){
    /* Schur complement computation */
    StrumpackDensePackage<float,float>* ptr=static_cast<StrumpackDensePackage<float,float>*>(SDP_C->SDP);
    SDP_C_float_copy_fields(SDP_C);
    ptr->compute_schur();
  }

  void SDP_C_float_reduce_RHS(StrumpackDensePackage_C_float *SDP_C, float *X, int *descX, float* B, int *descB){
    /* RHS reduction/condensation */
    StrumpackDensePackage<float,float>* ptr=static_cast<StrumpackDensePackage<float,float>*>(SDP_C->SDP);
    SDP_C_float_copy_fields(SDP_C);
    ptr->reduce_RHS(X,descX,B,descB);
  }

  void SDP_C_float_schur_product(StrumpackDensePackage_C_float *SDP_C, char Trans, float alpha, float *X, int *descX, float beta, float *B, int *descB){
    /* Schur complement product */
    StrumpackDensePackage<float,float>* ptr=static_cast<StrumpackDensePackage<float,float>*>(SDP_C->SDP);
    SDP_C_float_copy_fields(SDP_C);
    ptr->schur_product(Trans,alpha,X,descX,beta,B,descB);
  }

  void SDP_C_float_expand_solution(StrumpackDensePackage_C_float *SDP_C, float *X, int *descX, float *B, int *descB){
    /* Solution expansion */
    StrumpackDensePackage<float,float>* ptr=static_cast<StrumpackDensePackage<float,float>*>(SDP_C->SDP);
    SDP_C_float_copy_fields(SDP_C);
    ptr->expand_solution(X,descX,B,descB);
  }

  void SDP_C_float_solve(StrumpackDensePackage_C_float *SDP_C, float *X, int *descX, float *B, int *descB){
    /* Triangular solve */
    StrumpackDensePackage<float,float>* ptr=static_cast<StrumpackDensePackage<float,float>*>(SDP_C->SDP);
    SDP_C_float_copy_fields(SDP_C);
    ptr->solve(X,descX,B,descB);
  }

  float SDP_C_float_check_solution(StrumpackDensePackage_C_float *SDP_C, float *A, int *descA, float *X, int *descX, float *B, int *descB){
    /* Solution accuracy check */
    StrumpackDensePackage<float,float>* ptr=static_cast<StrumpackDensePackage<float,float>*>(SDP_C->SDP);
    SDP_C_float_copy_fields(SDP_C);
    return ptr->check_solution(A,descA,X,descX,B,descB);
  }

  float SDP_C_float_refine(StrumpackDensePackage_C_float *SDP_C, float *A, int *descA, float *X, int *descX, float *B, int *descB){
    /* Iterative refinement */
    StrumpackDensePackage<float,float>* ptr=static_cast<StrumpackDensePackage<float,float>*>(SDP_C->SDP);
    SDP_C_float_copy_fields(SDP_C);
    return ptr->refine(A,descA,X,descX,B,descB);
  }

  void SDP_C_float_product(StrumpackDensePackage_C_float *SDP_C, char Trans, float alpha, float *A, int *descA, float *B, int *descB, float beta, float *C, int *descC){
    /* Matrix-vector product */
    StrumpackDensePackage<float,float>* ptr=static_cast<StrumpackDensePackage<float,float>*>(SDP_C->SDP);
    SDP_C_float_copy_fields(SDP_C);
    ptr->product(Trans,alpha,A,descA,B,descB,beta,C,descC);
  }

  void SDP_C_float_extract(StrumpackDensePackage_C_float *SDP_C, float *A, int *descA, float *B, int *descB, int *I, int nI, int *J, int nJ){
    /* Element extraction */
    StrumpackDensePackage<float,float>* ptr=static_cast<StrumpackDensePackage<float,float>*>(SDP_C->SDP);
    SDP_C_float_copy_fields(SDP_C);
    ptr->extract(A,descA,B,descB,I,nI,J,nJ);
  }

  void SDP_C_float_extract_schur(StrumpackDensePackage_C_float *SDP_C, float *B, int *descB, int *I, int nI, int *J, int nJ){
    /* Schur complement element extraction */
    StrumpackDensePackage<float,float>* ptr=static_cast<StrumpackDensePackage<float,float>*>(SDP_C->SDP);
    SDP_C_float_copy_fields(SDP_C);
    ptr->extract_schur(B,descB,I,nI,J,nJ);
  }

  void SDP_C_float_print_statistics(StrumpackDensePackage_C_float *SDP_C){
    /* Statistics */
    StrumpackDensePackage<float,float>* ptr=static_cast<StrumpackDensePackage<float,float>*>(SDP_C->SDP);
    SDP_C_float_copy_fields(SDP_C);
    ptr->print_statistics();
  }

/*****************************************************************************/
  /* Complex double precision */
  void SDP_C_dcomplex_copy_fields(StrumpackDensePackage_C_dcomplex *SDP_C) {
    /* Copy all the fields from the C structure to the C++ class */
    StrumpackDensePackage<dcomplex,double>* ptr=static_cast<StrumpackDensePackage<dcomplex,double>*>(SDP_C->SDP);
    ptr->use_HSS         = SDP_C->use_HSS;
    ptr->tol_HSS         = SDP_C->tol_HSS;
    ptr->min_rand_HSS    = SDP_C->min_rand_HSS;
    ptr->max_rand_HSS    = SDP_C->max_rand_HSS;
    ptr->lim_rand_HSS    = SDP_C->lim_rand_HSS;
    ptr->inc_rand_HSS    = SDP_C->inc_rand_HSS;
    ptr->block_HSS       = SDP_C->block_HSS;
    ptr->levels_HSS      = SDP_C->levels_HSS;
    ptr->split_HSS       = SDP_C->split_HSS;
    ptr->tol_IR          = SDP_C->tol_IR;
    ptr->steps_IR        = SDP_C->steps_IR;
    ptr->fast_IR         = SDP_C->fast_IR;
    ptr->verbose         = SDP_C->verbose;
    ptr->access_is_global= SDP_C->access_is_global;
    ptr->sched_strat     = SDP_C->sched_strat;
    ptr->obj             = SDP_C->obj;
  }

  void SDP_C_dcomplex_init(StrumpackDensePackage_C_dcomplex *SDP_C, MPI_Comm comm) {
    /* Initialization: calls the C++ constructor */
    SDP_C->SDP=(void*)new StrumpackDensePackage<dcomplex,double>(comm);
    /* Copy default options to C structure */
    StrumpackDensePackage<dcomplex,double>* ptr=static_cast<StrumpackDensePackage<dcomplex,double>*>(SDP_C->SDP);
    SDP_C->use_HSS         = ptr->use_HSS;
    SDP_C->tol_HSS         = ptr->tol_HSS;
    SDP_C->min_rand_HSS    = ptr->min_rand_HSS;
    SDP_C->max_rand_HSS    = ptr->max_rand_HSS;
    SDP_C->lim_rand_HSS    = ptr->lim_rand_HSS;
    SDP_C->inc_rand_HSS    = ptr->inc_rand_HSS;
    SDP_C->block_HSS       = ptr->block_HSS;
    SDP_C->levels_HSS      = ptr->levels_HSS;
    SDP_C->split_HSS       = ptr->split_HSS;
    SDP_C->tol_IR          = ptr->tol_IR;
    SDP_C->steps_IR        = ptr->steps_IR;
    SDP_C->fast_IR         = ptr->fast_IR;
    SDP_C->verbose         = ptr->verbose;
    SDP_C->access_is_global= ptr->access_is_global;
    SDP_C->sched_strat     = ptr->sched_strat;
    SDP_C->obj             = ptr->obj;
  }

  void SDP_C_dcomplex_init_F90(StrumpackDensePackage_C_dcomplex *SDP_C, int fcomm) {
    /* Initialization: calls the C++ constructor */
    SDP_C->SDP=static_cast<void*>(new StrumpackDensePackage<dcomplex,double>(MPI_Comm_f2c(fcomm)));
    /* Copy default options to C structure */
    StrumpackDensePackage<dcomplex,double>* ptr=static_cast<StrumpackDensePackage<dcomplex,double>*>(SDP_C->SDP);
    SDP_C->use_HSS         = ptr->use_HSS;
    SDP_C->tol_HSS         = ptr->tol_HSS;
    SDP_C->min_rand_HSS    = ptr->min_rand_HSS;
    SDP_C->max_rand_HSS    = ptr->max_rand_HSS;
    SDP_C->lim_rand_HSS    = ptr->lim_rand_HSS;
    SDP_C->inc_rand_HSS    = ptr->inc_rand_HSS;
    SDP_C->block_HSS       = ptr->block_HSS;
    SDP_C->levels_HSS      = ptr->levels_HSS;
    SDP_C->split_HSS       = ptr->split_HSS;
    SDP_C->tol_IR          = ptr->tol_IR;
    SDP_C->steps_IR        = ptr->steps_IR;
    SDP_C->fast_IR         = ptr->fast_IR;
    SDP_C->verbose         = ptr->verbose;
    SDP_C->access_is_global= ptr->access_is_global;
    SDP_C->sched_strat     = ptr->sched_strat;
    SDP_C->obj             = ptr->obj;
  }

  void SDP_C_dcomplex_destroy(StrumpackDensePackage_C_dcomplex *SDP_C) {
    /* Cleanup: calls to C++ destructor */
    StrumpackDensePackage<dcomplex,double>* ptr=static_cast<StrumpackDensePackage<dcomplex,double>*>(SDP_C->SDP);
    delete ptr;
  }

  void SDP_C_dcomplex_compress_A(StrumpackDensePackage_C_dcomplex *SDP_C, dcomplex *A, int *descA) {
    /* Compression with an explicit matrix */
    StrumpackDensePackage<dcomplex,double>* ptr=static_cast<StrumpackDensePackage<dcomplex,double>*>(SDP_C->SDP);
    SDP_C_dcomplex_copy_fields(SDP_C);
    ptr->compress(A,descA);
  }

  void SDP_C_dcomplex_compress_ARS(StrumpackDensePackage_C_dcomplex *SDP_C, dcomplex* A, int *descA, dcomplex *Rr, dcomplex *Rc, dcomplex* Sr, dcomplex* Sc, int *descRS) {
    /* Compression with an explicit matrix and user-given samples and random vectors */
    StrumpackDensePackage<dcomplex,double>* ptr=static_cast<StrumpackDensePackage<dcomplex,double>*>(SDP_C->SDP);
    SDP_C_dcomplex_copy_fields(SDP_C);
    ptr->compress(A,descA,Rr,Rc,Sr,Sc,descRS);
  }

  void SDP_C_dcomplex_compress_mtxfree(StrumpackDensePackage_C_dcomplex *SDP_C, dcomplex *Rr, dcomplex*Rc, dcomplex*Sr, dcomplex* Sc, int *descRS, void (*elements)(void *, int *, int *, dcomplex *, int *)) {
    /* Matrix-free compression */
    StrumpackDensePackage<dcomplex,double>* ptr=static_cast<StrumpackDensePackage<dcomplex,double>*>(SDP_C->SDP);
    SDP_C_dcomplex_copy_fields(SDP_C);
    ptr->compress(Rr,Rc,Sr,Sc,descRS,elements);
  }

  double SDP_C_dcomplex_check_compression(StrumpackDensePackage_C_dcomplex *SDP_C, dcomplex *A, int *descA) {
    /* Compression quality check */
    StrumpackDensePackage<dcomplex,double>* ptr=static_cast<StrumpackDensePackage<dcomplex,double>*>(SDP_C->SDP);
    SDP_C_dcomplex_copy_fields(SDP_C);
    return ptr->check_compression(A,descA);
  }

  void SDP_C_dcomplex_factor(StrumpackDensePackage_C_dcomplex *SDP_C, dcomplex *A, int *descA) {
    /* Factorization */
    StrumpackDensePackage<dcomplex,double>* ptr=static_cast<StrumpackDensePackage<dcomplex,double>*>(SDP_C->SDP);
    SDP_C_dcomplex_copy_fields(SDP_C);
    ptr->factor(A,descA);
  }

  void SDP_C_dcomplex_partially_factor(StrumpackDensePackage_C_dcomplex *SDP_C, dcomplex* A, int *descA, int nfact) {
    /* Partial factorization */
    StrumpackDensePackage<dcomplex,double>* ptr=static_cast<StrumpackDensePackage<dcomplex,double>*>(SDP_C->SDP);
    SDP_C_dcomplex_copy_fields(SDP_C);
    ptr->partially_factor(A,descA,nfact);
  }

  void SDP_C_dcomplex_compute_schur(StrumpackDensePackage_C_dcomplex *SDP_C){
    /* Schur complement computation */
    StrumpackDensePackage<dcomplex,double>* ptr=static_cast<StrumpackDensePackage<dcomplex,double>*>(SDP_C->SDP);
    SDP_C_dcomplex_copy_fields(SDP_C);
    ptr->compute_schur();
  }

  void SDP_C_dcomplex_reduce_RHS(StrumpackDensePackage_C_dcomplex *SDP_C, dcomplex *X, int *descX, dcomplex* B, int *descB){
    /* RHS reduction/condensation */
    StrumpackDensePackage<dcomplex,double>* ptr=static_cast<StrumpackDensePackage<dcomplex,double>*>(SDP_C->SDP);
    SDP_C_dcomplex_copy_fields(SDP_C);
    ptr->reduce_RHS(X,descX,B,descB);
  }

  void SDP_C_dcomplex_schur_product(StrumpackDensePackage_C_dcomplex *SDP_C, char Trans, dcomplex alpha, dcomplex *X, int *descX, dcomplex beta, dcomplex *B, int *descB){
    /* Schur complement product */
    StrumpackDensePackage<dcomplex,double>* ptr=static_cast<StrumpackDensePackage<dcomplex,double>*>(SDP_C->SDP);
    SDP_C_dcomplex_copy_fields(SDP_C);
    ptr->schur_product(Trans,alpha,X,descX,beta,B,descB);
  }

  void SDP_C_dcomplex_expand_solution(StrumpackDensePackage_C_dcomplex *SDP_C, dcomplex *X, int *descX, dcomplex *B, int *descB){
    /* Solution expansion */
    StrumpackDensePackage<dcomplex,double>* ptr=static_cast<StrumpackDensePackage<dcomplex,double>*>(SDP_C->SDP);
    SDP_C_dcomplex_copy_fields(SDP_C);
    ptr->expand_solution(X,descX,B,descB);
  }

  void SDP_C_dcomplex_solve(StrumpackDensePackage_C_dcomplex *SDP_C, dcomplex *X, int *descX, dcomplex *B, int *descB){
    /* Triangular solve */
    StrumpackDensePackage<dcomplex,double>* ptr=static_cast<StrumpackDensePackage<dcomplex,double>*>(SDP_C->SDP);
    SDP_C_dcomplex_copy_fields(SDP_C);
    ptr->solve(X,descX,B,descB);
  }

  double SDP_C_dcomplex_check_solution(StrumpackDensePackage_C_dcomplex *SDP_C, dcomplex *A, int *descA, dcomplex *X, int *descX, dcomplex *B, int *descB){
    /* Solution accuracy check */
    StrumpackDensePackage<dcomplex,double>* ptr=static_cast<StrumpackDensePackage<dcomplex,double>*>(SDP_C->SDP);
    SDP_C_dcomplex_copy_fields(SDP_C);
    return ptr->check_solution(A,descA,X,descX,B,descB);
  }

  double SDP_C_dcomplex_refine(StrumpackDensePackage_C_dcomplex *SDP_C, dcomplex *A, int *descA, dcomplex *X, int *descX, dcomplex *B, int *descB){
    /* Iterative refinement */
    StrumpackDensePackage<dcomplex,double>* ptr=static_cast<StrumpackDensePackage<dcomplex,double>*>(SDP_C->SDP);
    SDP_C_dcomplex_copy_fields(SDP_C);
    return ptr->refine(A,descA,X,descX,B,descB);
  }

  void SDP_C_dcomplex_product(StrumpackDensePackage_C_dcomplex *SDP_C, char Trans, dcomplex alpha, dcomplex *A, int *descA, dcomplex *B, int *descB, dcomplex beta, dcomplex *C, int *descC){
    /* Matrix-vector product */
    StrumpackDensePackage<dcomplex,double>* ptr=static_cast<StrumpackDensePackage<dcomplex,double>*>(SDP_C->SDP);
    SDP_C_dcomplex_copy_fields(SDP_C);
    ptr->product(Trans,alpha,A,descA,B,descB,beta,C,descC);
  }

  void SDP_C_dcomplex_extract(StrumpackDensePackage_C_dcomplex *SDP_C, dcomplex *A, int *descA, dcomplex *B, int *descB, int *I, int nI, int *J, int nJ){
    /* Element extraction */
    StrumpackDensePackage<dcomplex,double>* ptr=static_cast<StrumpackDensePackage<dcomplex,double>*>(SDP_C->SDP);
    SDP_C_dcomplex_copy_fields(SDP_C);
    ptr->extract(A,descA,B,descB,I,nI,J,nJ);
  }

  void SDP_C_dcomplex_extract_schur(StrumpackDensePackage_C_dcomplex *SDP_C, dcomplex *B, int *descB, int *I, int nI, int *J, int nJ){
    /* Schur complement element extraction */
    StrumpackDensePackage<dcomplex,double>* ptr=static_cast<StrumpackDensePackage<dcomplex,double>*>(SDP_C->SDP);
    SDP_C_dcomplex_copy_fields(SDP_C);
    ptr->extract_schur(B,descB,I,nI,J,nJ);
  }

  void SDP_C_dcomplex_print_statistics(StrumpackDensePackage_C_dcomplex *SDP_C){
    /* Statistics */
    StrumpackDensePackage<dcomplex,double>* ptr=static_cast<StrumpackDensePackage<dcomplex,double>*>(SDP_C->SDP);
    SDP_C_dcomplex_copy_fields(SDP_C);
    ptr->print_statistics();
  }

/*****************************************************************************/
  /* Complex single precision */
  void SDP_C_scomplex_copy_fields(StrumpackDensePackage_C_scomplex *SDP_C) {
    /* Copy all the fields from the C structure to the C++ class */
    StrumpackDensePackage<scomplex,float>* ptr=static_cast<StrumpackDensePackage<scomplex,float>*>(SDP_C->SDP);
    ptr->use_HSS         = SDP_C->use_HSS;
    ptr->tol_HSS         = SDP_C->tol_HSS;
    ptr->min_rand_HSS    = SDP_C->min_rand_HSS;
    ptr->max_rand_HSS    = SDP_C->max_rand_HSS;
    ptr->lim_rand_HSS    = SDP_C->lim_rand_HSS;
    ptr->inc_rand_HSS    = SDP_C->inc_rand_HSS;
    ptr->block_HSS       = SDP_C->block_HSS;
    ptr->levels_HSS      = SDP_C->levels_HSS;
    ptr->split_HSS       = SDP_C->split_HSS;
    ptr->tol_IR          = SDP_C->tol_IR;
    ptr->steps_IR        = SDP_C->steps_IR;
    ptr->fast_IR         = SDP_C->fast_IR;
    ptr->verbose         = SDP_C->verbose;
    ptr->access_is_global= SDP_C->access_is_global;
    ptr->sched_strat     = SDP_C->sched_strat;
    ptr->obj             = SDP_C->obj;
  }

  void SDP_C_scomplex_init(StrumpackDensePackage_C_scomplex *SDP_C, MPI_Comm comm) {
    /* Initialization: calls the C++ constructor */
    SDP_C->SDP=(void*)new StrumpackDensePackage<scomplex,float>(comm);
    /* Copy default options to C structure */
    StrumpackDensePackage<scomplex,float>* ptr=static_cast<StrumpackDensePackage<scomplex,float>*>(SDP_C->SDP);
    SDP_C->use_HSS         = ptr->use_HSS;
    SDP_C->tol_HSS         = ptr->tol_HSS;
    SDP_C->min_rand_HSS    = ptr->min_rand_HSS;
    SDP_C->max_rand_HSS    = ptr->max_rand_HSS;
    SDP_C->lim_rand_HSS    = ptr->lim_rand_HSS;
    SDP_C->inc_rand_HSS    = ptr->inc_rand_HSS;
    SDP_C->block_HSS       = ptr->block_HSS;
    SDP_C->levels_HSS      = ptr->levels_HSS;
    SDP_C->split_HSS       = ptr->split_HSS;
    SDP_C->tol_IR          = ptr->tol_IR;
    SDP_C->steps_IR        = ptr->steps_IR;
    SDP_C->fast_IR         = ptr->fast_IR;
    SDP_C->verbose         = ptr->verbose;
    SDP_C->access_is_global= ptr->access_is_global;
    SDP_C->sched_strat     = ptr->sched_strat;
    SDP_C->obj             = ptr->obj;
  }

  void SDP_C_scomplex_destroy(StrumpackDensePackage_C_scomplex *SDP_C) {
    /* Cleanup: calls to C++ destructor */
    StrumpackDensePackage<scomplex,float>* ptr=static_cast<StrumpackDensePackage<scomplex,float>*>(SDP_C->SDP);
    delete ptr;
  }

  void SDP_C_scomplex_compress_A(StrumpackDensePackage_C_scomplex *SDP_C, scomplex *A, int *descA) {
    /* Compression with an explicit matrix */
    StrumpackDensePackage<scomplex,float>* ptr=static_cast<StrumpackDensePackage<scomplex,float>*>(SDP_C->SDP);
    SDP_C_scomplex_copy_fields(SDP_C);
    ptr->compress(A,descA);
  }

  void SDP_C_scomplex_compress_ARS(StrumpackDensePackage_C_scomplex *SDP_C, scomplex* A, int *descA, scomplex *Rr, scomplex *Rc, scomplex* Sr, scomplex* Sc, int *descRS) {
    /* Compression with an explicit matrix and user-given samples and random vectors */
    StrumpackDensePackage<scomplex,float>* ptr=static_cast<StrumpackDensePackage<scomplex,float>*>(SDP_C->SDP);
    SDP_C_scomplex_copy_fields(SDP_C);
    ptr->compress(A,descA,Rr,Rc,Sr,Sc,descRS);
  }

  void SDP_C_scomplex_compress_mtxfree(StrumpackDensePackage_C_scomplex *SDP_C, scomplex *Rr, scomplex*Rc, scomplex*Sr, scomplex* Sc, int *descRS, void (*elements)(void *, int *, int *, scomplex *, int *)) {
    /* Matrix-free compression */
    StrumpackDensePackage<scomplex,float>* ptr=static_cast<StrumpackDensePackage<scomplex,float>*>(SDP_C->SDP);
    SDP_C_scomplex_copy_fields(SDP_C);
    ptr->compress(Rr,Rc,Sr,Sc,descRS,elements);
  }

  float SDP_C_scomplex_check_compression(StrumpackDensePackage_C_scomplex *SDP_C, scomplex *A, int *descA) {
    /* Compression quality check */
    StrumpackDensePackage<scomplex,float>* ptr=static_cast<StrumpackDensePackage<scomplex,float>*>(SDP_C->SDP);
    SDP_C_scomplex_copy_fields(SDP_C);
    return ptr->check_compression(A,descA);
  }

  void SDP_C_scomplex_factor(StrumpackDensePackage_C_scomplex *SDP_C, scomplex *A, int *descA) {
    /* Factorization */
    StrumpackDensePackage<scomplex,float>* ptr=static_cast<StrumpackDensePackage<scomplex,float>*>(SDP_C->SDP);
    SDP_C_scomplex_copy_fields(SDP_C);
    ptr->factor(A,descA);
  }

  void SDP_C_scomplex_partially_factor(StrumpackDensePackage_C_scomplex *SDP_C, scomplex* A, int *descA, int nfact) {
    /* Partial factorization */
    StrumpackDensePackage<scomplex,float>* ptr=static_cast<StrumpackDensePackage<scomplex,float>*>(SDP_C->SDP);
    SDP_C_scomplex_copy_fields(SDP_C);
    ptr->partially_factor(A,descA,nfact);
  }

  void SDP_C_scomplex_compute_schur(StrumpackDensePackage_C_scomplex *SDP_C){
    /* Schur complement computation */
    StrumpackDensePackage<scomplex,float>* ptr=static_cast<StrumpackDensePackage<scomplex,float>*>(SDP_C->SDP);
    SDP_C_scomplex_copy_fields(SDP_C);
    ptr->compute_schur();
  }

  void SDP_C_scomplex_reduce_RHS(StrumpackDensePackage_C_scomplex *SDP_C, scomplex *X, int *descX, scomplex* B, int *descB){
    /* RHS reduction/condensation */
    StrumpackDensePackage<scomplex,float>* ptr=static_cast<StrumpackDensePackage<scomplex,float>*>(SDP_C->SDP);
    SDP_C_scomplex_copy_fields(SDP_C);
    ptr->reduce_RHS(X,descX,B,descB);
  }

  void SDP_C_scomplex_schur_product(StrumpackDensePackage_C_scomplex *SDP_C, char Trans, scomplex alpha, scomplex *X, int *descX, scomplex beta, scomplex *B, int *descB){
    /* Schur complement product */
    StrumpackDensePackage<scomplex,float>* ptr=static_cast<StrumpackDensePackage<scomplex,float>*>(SDP_C->SDP);
    SDP_C_scomplex_copy_fields(SDP_C);
    ptr->schur_product(Trans,alpha,X,descX,beta,B,descB);
  }

  void SDP_C_scomplex_expand_solution(StrumpackDensePackage_C_scomplex *SDP_C, scomplex *X, int *descX, scomplex *B, int *descB){
    /* Solution expansion */
    StrumpackDensePackage<scomplex,float>* ptr=static_cast<StrumpackDensePackage<scomplex,float>*>(SDP_C->SDP);
    SDP_C_scomplex_copy_fields(SDP_C);
    ptr->expand_solution(X,descX,B,descB);
  }

  void SDP_C_scomplex_solve(StrumpackDensePackage_C_scomplex *SDP_C, scomplex *X, int *descX, scomplex *B, int *descB){
    /* Triangular solve */
    StrumpackDensePackage<scomplex,float>* ptr=static_cast<StrumpackDensePackage<scomplex,float>*>(SDP_C->SDP);
    SDP_C_scomplex_copy_fields(SDP_C);
    ptr->solve(X,descX,B,descB);
  }

  float SDP_C_scomplex_check_solution(StrumpackDensePackage_C_scomplex *SDP_C, scomplex *A, int *descA, scomplex *X, int *descX, scomplex *B, int *descB){
    /* Solution accuracy check */
    StrumpackDensePackage<scomplex,float>* ptr=static_cast<StrumpackDensePackage<scomplex,float>*>(SDP_C->SDP);
    SDP_C_scomplex_copy_fields(SDP_C);
    return ptr->check_solution(A,descA,X,descX,B,descB);
  }

  float SDP_C_scomplex_refine(StrumpackDensePackage_C_scomplex *SDP_C, scomplex *A, int *descA, scomplex *X, int *descX, scomplex *B, int *descB){
    /* Iterative refinement */
    StrumpackDensePackage<scomplex,float>* ptr=static_cast<StrumpackDensePackage<scomplex,float>*>(SDP_C->SDP);
    SDP_C_scomplex_copy_fields(SDP_C);
    return ptr->refine(A,descA,X,descX,B,descB);
  }

  void SDP_C_scomplex_product(StrumpackDensePackage_C_scomplex *SDP_C, char Trans, scomplex alpha, scomplex *A, int *descA, scomplex *B, int *descB, scomplex beta, scomplex *C, int *descC){
    /* Matrix-vector product */
    StrumpackDensePackage<scomplex,float>* ptr=static_cast<StrumpackDensePackage<scomplex,float>*>(SDP_C->SDP);
    SDP_C_scomplex_copy_fields(SDP_C);
    ptr->product(Trans,alpha,A,descA,B,descB,beta,C,descC);
  }

  void SDP_C_scomplex_extract(StrumpackDensePackage_C_scomplex *SDP_C, scomplex *A, int *descA, scomplex *B, int *descB, int *I, int nI, int *J, int nJ){
    /* Element extraction */
    StrumpackDensePackage<scomplex,float>* ptr=static_cast<StrumpackDensePackage<scomplex,float>*>(SDP_C->SDP);
    SDP_C_scomplex_copy_fields(SDP_C);
    ptr->extract(A,descA,B,descB,I,nI,J,nJ);
  }

  void SDP_C_scomplex_extract_schur(StrumpackDensePackage_C_scomplex *SDP_C, scomplex *B, int *descB, int *I, int nI, int *J, int nJ){
    /* Schur complement element extraction */
    StrumpackDensePackage<scomplex,float>* ptr=static_cast<StrumpackDensePackage<scomplex,float>*>(SDP_C->SDP);
    SDP_C_scomplex_copy_fields(SDP_C);
    ptr->extract_schur(B,descB,I,nI,J,nJ);
  }

  void SDP_C_scomplex_print_statistics(StrumpackDensePackage_C_scomplex *SDP_C){
    /* Statistics */
    StrumpackDensePackage<scomplex,float>* ptr=static_cast<StrumpackDensePackage<scomplex,float>*>(SDP_C->SDP);
    SDP_C_scomplex_copy_fields(SDP_C);
    ptr->print_statistics();
  }

}
