/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */

#ifndef STRUMPACKRAND_HPP
#define STRUMPACKRAND_HPP

class StrumpackRand {

private:

#if defined(RANDGEN)
  /* C++11 random number generators */
  std::mt19937 *mt_rand;
#if defined(RANDGENNORMAL)
  /* Normal distribution */
  std::normal_distribution<double> *d_rand;
#else
  /* Uniform distribution (default) */
  std::uniform_real_distribution<double> *d_rand;
#endif
#endif

public:

  StrumpackRand(int s) {
#if defined(RANDGEN)
    mt_rand=new std::mt19937(s);
#if defined(RANDGENNORMAL)
    d_rand=new std::normal_distribution<double>(0.0,1.0);
#else
    d_rand=new std::uniform_real_distribution<double>(0.0,1.0);
#endif
#else
    srand(s);
#endif
  }

  double operator()() {
#if defined(RANDGEN)
    return (*d_rand)(*mt_rand);
#else
    return static_cast<double>(rand())/static_cast<double>(RAND_MAX);
#endif

  }
};

#endif // STRUMPACKRAND_HPP
