/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */

#ifndef LAPACK_H
#define LAPACK_H

/* LAPACK routines */

  /* xLACPY */
  int dlacpy_(char *, int *, int *,   double *, int *,   double *, int *);
  int slacpy_(char *, int *, int *,    float *, int *,    float *, int *);
  int zlacpy_(char *, int *, int *, dcomplex *, int *, dcomplex *, int *);
  int clacpy_(char *, int *, int *, scomplex *, int *, scomplex *, int *);

  /* xLAPMT */
  void dlapmt_(int *, int *, int *,   double *, int *, int *);
  void slapmt_(int *, int *, int *,    float *, int *, int *);
  void zlapmt_(int *, int *, int *, dcomplex *, int *, int *);
  void clapmt_(int *, int *, int *, scomplex *, int *, int *);

  /* xLANGE */
  double dlange_(char *,int *,int *,   double *,int *, double *);
   float slange_(char *,int *,int *,    float *,int *,  float *);
  double zlange_(char *,int *,int *, dcomplex *,int *, double *);
   float clange_(char *,int *,int *, scomplex *,int *,  float *);

  /* xLACGV */
  void zlacgv_(int *, dcomplex *, int *);
  void clacgv_(int *, scomplex *, int *);

  /* xGETRF */
  void dgetrf_(int *, int *,   double *, int *, int *, int *);
  void sgetrf_(int *, int *,    float *, int *, int *, int *);
  void zgetrf_(int *, int *, dcomplex *, int *, int *, int *);
  void cgetrf_(int *, int *, scomplex *, int *, int *, int *);

  /* xGETRS */
  void dgetrs_(char *, int *, int *,   double *, int *, int *,   double *, int *, int *);
  void sgetrs_(char *, int *, int *,    float *, int *, int *,    float *, int *, int *);
  void zgetrs_(char *, int *, int *, dcomplex *, int *, int *, dcomplex *, int *, int *);
  void cgetrs_(char *, int *, int *, scomplex *, int *, int *, scomplex *, int *, int *);

  /* xGELQF */
  void dgelqf_(int *, int *,   double *, int *,   double *,   double *, int *, int *);
  void sgelqf_(int *, int *,    float *, int *,    float *,    float *, int *, int *);
  void zgelqf_(int *, int *, dcomplex *, int *, dcomplex *, dcomplex *, int *, int *);
  void cgelqf_(int *, int *, scomplex *, int *, scomplex *, scomplex *, int *, int *);

  /* xxxGLQ */
  void dorglq_(int *, int *, int *,   double *, int *,   double *,   double *, int *, int *);
  void sorglq_(int *, int *, int *,    float *, int *,    float *,    float *, int *, int *);
  void zunglq_(int *, int *, int *, dcomplex *, int *, dcomplex *, dcomplex *, int *, int *);
  void cunglq_(int *, int *, int *, scomplex *, int *, scomplex *, scomplex *, int *, int *);

  /* xGEQP3mod */
    void dgeqp3mod_(int *, int *, double *, int *, int *,   double *, double *, int *, int *, int *, double *);
    void sgeqp3mod_(int *, int *,  float *, int *, int *,    float *,  float *, int *, int *, int *,  float *);
    void zgeqp3mod_(int *, int *, dcomplex *, int *, int *, dcomplex *, dcomplex *, int *, double *, int *, int *, double *);
    void cgeqp3mod_(int *, int *, scomplex *, int *, int *, scomplex *, scomplex *, int *, float *, int *, int *, float *);

  /* xLASWP */
  void dlaswp_(int *,   double *, int *, int *, int *, int *, int *);
  void slaswp_(int *,    float *, int *, int *, int *, int *, int *);
  void zlaswp_(int *, dcomplex *, int *, int *, int *, int *, int *);
  void claswp_(int *, scomplex *, int *, int *, int *, int *, int *);

#endif // LAPACK_H
