/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */

#ifndef SCALAPACK_HPP
#define SCALAPACK_HPP

/* ScaLAPACK routines */
  /* Arithmetic-independent routines */
  inline int descinit(int* desc, int m, int n, int mb, int nb, int rsrc, int csrc, int ictxt, int mxllda) {
    int info;
    descinit_(desc, &m, &n, &mb, &nb, &rsrc, &csrc, &ictxt, &mxllda, &info);
    return info;
  }
  inline void descset(int* desc, int m, int n, int mb, int nb, int rsrc, int csrc, int ictxt, int mxllda) {
    descset_(desc, &m, &n, &mb, &nb, &rsrc, &csrc, &ictxt, &mxllda);
  }

  inline int numroc(int n, int nb, int iproc, int isrcproc, int nprocs) {
    return numroc_(&n, &nb, &iproc, &isrcproc, &nprocs);
  }

  /* xGSUM2D */
  template<typename S> inline void gsum2d(int, char, char, int, int, S *, int, int, int);
  template<> inline void gsum2d<double>(int ctxt, char scope, char top, int m, int n, double *A, int lda, int rdest, int cdest) {
    dgsum2d_(&ctxt,&scope,&top,&m,&n,A,&lda,&rdest,&cdest);
  }
  template<> inline void gsum2d< float>(int ctxt, char scope, char top, int m, int n,  float *A, int lda, int rdest, int cdest) {
    sgsum2d_(&ctxt,&scope,&top,&m,&n,A,&lda,&rdest,&cdest);
  }

  /* xGAMX2D */
  template<typename S> inline void gamx2d(int, char, char, int, int, S *, int, int *, int *, int, int, int);
  template<> inline void gamx2d<double>(int ctxt, char scope, char top, int m, int n, double *A, int lda, int *ra, int *ca, int ldia, int rdest, int cdest) {
    dgamx2d_(&ctxt,&scope,&top,&m,&n,A,&lda,ra,ca,&ldia,&rdest,&cdest);
  }
  template<> inline void gamx2d< float>(int ctxt, char scope, char top, int m, int n,  float *A, int lda, int *ra, int *ca, int ldia, int rdest, int cdest) {
    sgamx2d_(&ctxt,&scope,&top,&m,&n,A,&lda,ra,ca,&ldia,&rdest,&cdest);
  }

  /* PxSWAP */
  template<typename T> inline void pswap(int, T *, int, int, int *, int, T *, int, int, int *, int);
  template<> inline void pswap<  double>(int n,   double *X, int ix, int jx, int *descX, int incx,   double *Y, int iy, int jy, int *descY, int incy) {
    pdswap_(&n,X,&ix,&jx,descX,&incx,Y,&iy,&jy,descY,&incy);
  }
  template<> inline void pswap<   float>(int n,    float *X, int ix, int jx, int *descX, int incx,    float *Y, int iy, int jy, int *descY, int incy) {
    psswap_(&n,X,&ix,&jx,descX,&incx,Y,&iy,&jy,descY,&incy);
  }
  template<> inline void pswap<dcomplex>(int n, dcomplex *X, int ix, int jx, int *descX, int incx, dcomplex *Y, int iy, int jy, int *descY, int incy) {
    pzswap_(&n,X,&ix,&jx,descX,&incx,Y,&iy,&jy,descY,&incy);
  }
  template<> inline void pswap<scomplex>(int n, scomplex *X, int ix, int jx, int *descX, int incx, scomplex *Y, int iy, int jy, int *descY, int incy) {
    pcswap_(&n,X,&ix,&jx,descX,&incx,Y,&iy,&jy,descY,&incy);
  }

  /* PxSCAL */
  template<typename T> inline void pscal(int, T, T *, int, int, int *, int);
  template<> inline void pscal<  double>(int n,   double a,   double *X, int ix, int jx, int *descX, int incx) {
    pdscal_(&n,&a,X,&ix,&jx,descX,&incx);
  }
  template<> inline void pscal<   float>(int n,    float a,    float *X, int ix, int jx, int *descX, int incx) {
    psscal_(&n,&a,X,&ix,&jx,descX,&incx);
  }
  template<> inline void pscal<dcomplex>(int n, dcomplex a, dcomplex *X, int ix, int jx, int *descX, int incx) {
    pzscal_(&n,&a,X,&ix,&jx,descX,&incx);
  }
  template<> inline void pscal<scomplex>(int n, scomplex a, scomplex *X, int ix, int jx, int *descX, int incx) {
    pcscal_(&n,&a,X,&ix,&jx,descX,&incx);
  }

  /* PxGEMV */
  template<typename T> inline void pgemv(char, int, int, T, T *, int, int, int *, T *, int, int, int*, int, T, T *, int, int, int *, int);
  template<> inline void pgemv<  double>(char transA, int m, int n,   double alpha,   double *A, int ia, int ja, int *descA,   double *X, int ix, int jx, int *descX, int incx,   double beta,   double *Y, int iy, int jy, int *descY, int incy) {
    pdgemv_(&transA,&m,&n,&alpha,A,&ia,&ja,descA,X,&ix,&jx,descX,&incx,&beta,Y,&iy,&jy,descY,&incy);
  }
  template<> inline void pgemv<   float>(char transA, int m, int n,    float alpha,    float *A, int ia, int ja, int *descA,    float *X, int ix, int jx, int *descX, int incx,    float beta,    float *Y, int iy, int jy, int *descY, int incy) {
    psgemv_(&transA,&m,&n,&alpha,A,&ia,&ja,descA,X,&ix,&jx,descX,&incx,&beta,Y,&iy,&jy,descY,&incy);
  }
  template<> inline void pgemv<dcomplex>(char transA, int m, int n, dcomplex alpha, dcomplex *A, int ia, int ja, int *descA, dcomplex *X, int ix, int jx, int *descX, int incx, dcomplex beta, dcomplex *Y, int iy, int jy, int *descY, int incy) {
    pzgemv_(&transA,&m,&n,&alpha,A,&ia,&ja,descA,X,&ix,&jx,descX,&incx,&beta,Y,&iy,&jy,descY,&incy);
  }
  template<> inline void pgemv<scomplex>(char transA, int m, int n, scomplex alpha, scomplex *A, int ia, int ja, int *descA, scomplex *X, int ix, int jx, int *descX, int incx, scomplex beta, scomplex *Y, int iy, int jy, int *descY, int incy) {
    pcgemv_(&transA,&m,&n,&alpha,A,&ia,&ja,descA,X,&ix,&jx,descX,&incx,&beta,Y,&iy,&jy,descY,&incy);
  }

  /* PxGEMM */
  template<typename T> inline void pgemm(char, char, int, int, int, T, T*, int, int, int*, T*, int, int, int*, T, T*, int, int, int*);
  template<> inline void pgemm<double> (char transA, char transB, int m, int n, int k,    double alpha,   double *A, int ia, int ja, int *descA,   double *B, int ib, int jb, int *descB,   double beta, double *C, int ic, int jc, int *descC) {
    pdgemm_(&transA,&transB,&m,&n,&k,&alpha,A,&ia,&ja,descA,B,&ib,&jb,descB,&beta,C,&ic,&jc,descC);
  }
  template<> inline void pgemm<float>  (char transA, char transB, int m, int n, int k,     float alpha,    float *A, int ia, int ja, int *descA,    float *B, int ib, int jb, int *descB,    float beta,   float *C, int ic, int jc, int *descC) {
    psgemm_(&transA,&transB,&m,&n,&k,&alpha,A,&ia,&ja,descA,B,&ib,&jb,descB,&beta,C,&ic,&jc,descC);
  }
  template<> inline void pgemm<dcomplex>(char transA, char transB, int m, int n, int k, dcomplex alpha, dcomplex *A, int ia, int ja, int *descA, dcomplex *B, int ib, int jb, int *descB, dcomplex beta, dcomplex *C, int ic, int jc, int *descC) {
    pzgemm_(&transA,&transB,&m,&n,&k,&alpha,A,&ia,&ja,descA,B,&ib,&jb,descB,&beta,C,&ic,&jc,descC);
  }
  template<> inline void pgemm<scomplex>(char transA, char transB, int m, int n, int k, scomplex alpha, scomplex *A, int ia, int ja, int *descA, scomplex *B, int ib, int jb, int *descB, scomplex beta, scomplex *C, int ic, int jc, int *descC) {
    pcgemm_(&transA,&transB,&m,&n,&k,&alpha,A,&ia,&ja,descA,B,&ib,&jb,descB,&beta,C,&ic,&jc,descC);
  }

  /* PxLACGV */
  template<typename T> inline void placgv(int, T *, int, int, int *, int);
  template<> inline void placgv<  double>(int n,   double *X, int ix, int jx, int *descX, int incx) {
    // Nothing to do
  }
  template<> inline void placgv<   float>(int n,    float *X, int ix, int jx, int *descX, int incx) {
    // Nothing to do
  }
  template<> inline void placgv<dcomplex>(int n, dcomplex *X, int ix, int jx, int *descX, int incx) {
    pzlacgv_(&n,X,&ix,&jx,descX,&incx);
  }
  template<> inline void placgv<scomplex>(int n, scomplex *X, int ix, int jx, int *descX, int incx) {
    pclacgv_(&n,X,&ix,&jx,descX,&incx);
  }

  /* PxGER */
  template<typename T> inline void pgeru(int, int, T, T*, int, int, int*, int, T*, int, int, int *, int, T*, int, int, int*);
  template<> inline void pgeru<  double>(int m, int n,   double alpha,   double *X, int ix, int jx, int *descX, int incx,   double *Y, int iy, int jy, int *descY, int incy,   double *A, int ia, int ja, int *descA) {
    pdger_ (&m,&n,&alpha,X,&ix,&jx,descX,&incx,Y,&iy,&jy,descY,&incy,A,&ia,&ja,descA);
  }
  template<> inline void pgeru<   float>(int m, int n,    float alpha,    float *X, int ix, int jx, int *descX, int incx,    float *Y, int iy, int jy, int *descY, int incy,    float *A, int ia, int ja, int *descA) {
    psger_ (&m,&n,&alpha,X,&ix,&jx,descX,&incx,Y,&iy,&jy,descY,&incy,A,&ia,&ja,descA);
  }
  template<> inline void pgeru<dcomplex>(int m, int n, dcomplex alpha, dcomplex *X, int ix, int jx, int *descX, int incx, dcomplex *Y, int iy, int jy, int *descY, int incy, dcomplex *A, int ia, int ja, int *descA) {
    pzgeru_(&m,&n,&alpha,X,&ix,&jx,descX,&incx,Y,&iy,&jy,descY,&incy,A,&ia,&ja,descA);
  }
  template<> inline void pgeru<scomplex>(int m, int n, scomplex alpha, scomplex *X, int ix, int jx, int *descX, int incx, scomplex *Y, int iy, int jy, int *descY, int incy, scomplex *A, int ia, int ja, int *descA) {
    pcgeru_(&m,&n,&alpha,X,&ix,&jx,descX,&incx,Y,&iy,&jy,descY,&incy,A,&ia,&ja,descA);
  }

  /* PxLASWP */
  template<typename scalar> inline void plaswp(char direc, char rowcol, int n, scalar* a, int ia, int ja, int* desca, int k1, int k2, int* ipiv);
  template<> inline void plaswp(char direc, char rowcol, int n, double* a, int ia, int ja, int* desca, int k1, int k2, int* ipiv) {
    pdlaswp_(&direc,&rowcol,&n,a,&ia,&ja,desca,&k1,&k2,ipiv);
  }
  template<> inline void plaswp(char direc, char rowcol, int n, float* a, int ia, int ja, int* desca, int k1, int k2, int* ipiv) {
    pslaswp_(&direc,&rowcol,&n,a,&ia,&ja,desca,&k1,&k2,ipiv);
  }
  template<> inline void plaswp(char direc, char rowcol, int n, dcomplex* a, int ia, int ja, int* desca, int k1, int k2, int* ipiv) {
    pzlaswp_(&direc,&rowcol,&n,a,&ia,&ja,desca,&k1,&k2,ipiv);
  }
  template<> inline void plaswp(char direc, char rowcol, int n, scomplex* a, int ia, int ja, int* desca, int k1, int k2, int* ipiv) {
    pclaswp_(&direc,&rowcol,&n,a,&ia,&ja,desca,&k1,&k2,ipiv);
  }

  /* PxLAPIV */
  template<typename scalar> inline void plapiv(char direc, char rowcol, char pivroc, int m, int n, scalar* a, int ia, int ja, int* desca, int* ipiv, int ip, int jp, int* descip, int* iwork);
  template<> inline void plapiv(char direc, char rowcol, char pivroc, int m, int n, double* a, int ia, int ja, int* desca, int* ipiv, int ip, int jp, int* descip, int* iwork) {
    pdlapiv_(&direc,&rowcol,&pivroc,&m,&n,a,&ia,&ja,desca,ipiv,&ip,&jp,descip,iwork);
  }
  template<> inline void plapiv(char direc, char rowcol, char pivroc, int m, int n, float* a, int ia, int ja, int* desca, int* ipiv, int ip, int jp, int* descip, int* iwork) {
    pslapiv_(&direc,&rowcol,&pivroc,&m,&n,a,&ia,&ja,desca,ipiv,&ip,&jp,descip,iwork);
  }
  template<> inline void plapiv(char direc, char rowcol, char pivroc, int m, int n, dcomplex* a, int ia, int ja, int* desca, int* ipiv, int ip, int jp, int* descip, int* iwork) {
    pzlapiv_(&direc,&rowcol,&pivroc,&m,&n,a,&ia,&ja,desca,ipiv,&ip,&jp,descip,iwork);
  }
  template<> inline void plapiv(char direc, char rowcol, char pivroc, int m, int n, scomplex* a, int ia, int ja, int* desca, int* ipiv, int ip, int jp, int* descip, int* iwork) {
    pclapiv_(&direc,&rowcol,&pivroc,&m,&n,a,&ia,&ja,desca,ipiv,&ip,&jp,descip,iwork);
  }

  /* PxTRSM */
  template<typename T> inline void ptrsm(char, char, char, char, int, int, T, T *, int, int , int *, T *, int ,int, int *);
  template<> inline void ptrsm<  double>(char side, char uplo, char trans, char diag, int m, int n,   double alpha,   double *A, int ia, int ja, int *descA,   double *B, int ib, int jb, int *descB) {
    pdtrsm_(&side,&uplo,&trans,&diag,&m,&n,&alpha,A,&ia,&ja,descA,B,&ib,&jb,descB);
  }
  template<> inline void ptrsm<   float>(char side, char uplo, char trans, char diag, int m, int n,    float alpha,    float *A, int ia, int ja, int *descA,    float *B, int ib, int jb, int *descB) {
    pstrsm_(&side,&uplo,&trans,&diag,&m,&n,&alpha,A,&ia,&ja,descA,B,&ib,&jb,descB);
  }
  template<> inline void ptrsm<dcomplex>(char side, char uplo, char trans, char diag, int m, int n, dcomplex alpha, dcomplex *A, int ia, int ja, int *descA, dcomplex *B, int ib, int jb, int *descB) {
    pztrsm_(&side,&uplo,&trans,&diag,&m,&n,&alpha,A,&ia,&ja,descA,B,&ib,&jb,descB);
  }
  template<> inline void ptrsm<scomplex>(char side, char uplo, char trans, char diag, int m, int n, scomplex alpha, scomplex *A, int ia, int ja, int *descA, scomplex *B, int ib, int jb, int *descB) {
    pctrsm_(&side,&uplo,&trans,&diag,&m,&n,&alpha,A,&ia,&ja,descA,B,&ib,&jb,descB);
  }

  /* PxTRSV */
  template<typename T> inline void ptrsv(char, char, char, int, T *, int, int , int *, T *, int ,int, int *, int);
  template<> inline void ptrsv<  double>(char uplo, char trans, char diag, int m, double *A, int ia, int ja, int *descA,   double *B, int ib, int jb, int *descB, int incB) {
    pdtrsv_(&uplo,&trans,&diag,&m,A,&ia,&ja,descA,B,&ib,&jb,descB,&incB);
  }
  template<> inline void ptrsv<   float>(char uplo, char trans, char diag, int m, float *A, int ia, int ja, int *descA,    float *B, int ib, int jb, int *descB, int incB) {
    pstrsv_(&uplo,&trans,&diag,&m,A,&ia,&ja,descA,B,&ib,&jb,descB,&incB);
  }
  template<> inline void ptrsv<dcomplex>(char uplo, char trans, char diag, int m, dcomplex *A, int ia, int ja, int *descA, dcomplex *B, int ib, int jb, int *descB, int incB) {
    pztrsv_(&uplo,&trans,&diag,&m,A,&ia,&ja,descA,B,&ib,&jb,descB,&incB);
  }
  template<> inline void ptrsv<scomplex>(char uplo, char trans, char diag, int m, scomplex *A, int ia, int ja, int *descA, scomplex *B, int ib, int jb, int *descB, int incB) {
    pctrsv_(&uplo,&trans,&diag,&m,A,&ia,&ja,descA,B,&ib,&jb,descB,&incB);
  }

  /* PxLANGE */
  template<typename T, typename S> inline S plange(char , int , int , T *, int, int, int *, S *);
  template<> inline double plange<  double,double>(char norm, int m, int n,   double *A, int ia, int ja, int *descA, double *work) {
    return pdlange_(&norm,&m,&n,A,&ia,&ja,descA,work);
  }
  template<> inline  float plange<   float, float>(char norm, int m, int n,    float *A, int ia, int ja, int *descA,  float *work) {
    return pslange_(&norm,&m,&n,A,&ia,&ja,descA,work);
  }
  template<> inline double plange<dcomplex,double>(char norm, int m, int n, dcomplex *A, int ia, int ja, int *descA, double *work) {
    return pzlange_(&norm,&m,&n,A,&ia,&ja,descA,work);
  }
  template<> inline  float plange<scomplex, float>(char norm, int m, int n, scomplex *A, int ia, int ja, int *descA,  float *work) {
    return pclange_(&norm,&m,&n,A,&ia,&ja,descA,work);
}

  /* PxGEADD */
  template<typename T> inline void pgeadd(char, int, int, T, T *, int, int, int *, T, T *, int, int, int *);
  template<> inline void pgeadd<  double>(char trans, int m, int n,   double alpha,   double *A, int ia, int ja, int *descA,   double beta,   double *C, int ic, int jc, int *descC) {
    pdgeadd_(&trans,&m,&n,&alpha,A,&ia,&ja,descA,&beta,C,&ic,&jc,descC);
  }
  template<> inline void pgeadd<   float>(char trans, int m, int n,    float alpha,    float *A, int ia, int ja, int *descA,    float beta,    float *C, int ic, int jc, int *descC) {
    psgeadd_(&trans,&m,&n,&alpha,A,&ia,&ja,descA,&beta,C,&ic,&jc,descC);
  }
  template<> inline void pgeadd<dcomplex>(char trans, int m, int n, dcomplex alpha, dcomplex *A, int ia, int ja, int *descA, dcomplex beta, dcomplex *C, int ic, int jc, int *descC) {
    pzgeadd_(&trans,&m,&n,&alpha,A,&ia,&ja,descA,&beta,C,&ic,&jc,descC);
  }
  template<> inline void pgeadd<scomplex>(char trans, int m, int n, scomplex alpha, scomplex *A, int ia, int ja, int *descA, scomplex beta, scomplex *C, int ic, int jc, int *descC) {
    pcgeadd_(&trans,&m,&n,&alpha,A,&ia,&ja,descA,&beta,C,&ic,&jc,descC);
  }

  /* PxLACPY */
  template<typename T> inline void placpy(char, int, int, T *, int, int, int *, T *, int, int, int *);
  template<> inline void placpy<  double>(char trans, int m, int n,   double *A, int ia, int ja, int *descA,   double *C, int ic, int jc, int *descC) {
    pdlacpy_(&trans,&m,&n,A,&ia,&ja,descA,C,&ic,&jc,descC);
  }
  template<> inline void placpy<   float>(char trans, int m, int n,    float *A, int ia, int ja, int *descA,    float *C, int ic, int jc, int *descC) {
    pslacpy_(&trans,&m,&n,A,&ia,&ja,descA,C,&ic,&jc,descC);
  }
  template<> inline void placpy<dcomplex>(char trans, int m, int n, dcomplex *A, int ia, int ja, int *descA, dcomplex *C, int ic, int jc, int *descC) {
    pzlacpy_(&trans,&m,&n,A,&ia,&ja,descA,C,&ic,&jc,descC);
  }
  template<> inline void placpy<scomplex>(char trans, int m, int n, scomplex *A, int ia, int ja, int *descA, scomplex *C, int ic, int jc, int *descC) {
    pclacpy_(&trans,&m,&n,A,&ia,&ja,descA,C,&ic,&jc,descC);
  }

  /* pxGEMR2D */
  template<typename T> inline void pgemr2d(int, int, T *, int, int, int *, T *, int, int, int *, int);
  template<> inline void pgemr2d<  double>(int m, int n,   double *A, int ia, int ja, int *descA,   double *B, int ib, int jb, int *descB, int ctxt) {
    pdgemr2d_(&m,&n,A,&ia,&ja,descA,B,&ib,&jb,descB,&ctxt);
  }
  template<> inline void pgemr2d<   float>(int m, int n,    float *A, int ia, int ja, int *descA,    float *B, int ib, int jb, int *descB, int ctxt) {
    psgemr2d_(&m,&n,A,&ia,&ja,descA,B,&ib,&jb,descB,&ctxt);
  }
  template<> inline void pgemr2d<dcomplex>(int m, int n, dcomplex *A, int ia, int ja, int *descA, dcomplex *B, int ib, int jb, int *descB, int ctxt) {
    pzgemr2d_(&m,&n,A,&ia,&ja,descA,B,&ib,&jb,descB,&ctxt);
  }
  template<> inline void pgemr2d<scomplex>(int m, int n, scomplex *A, int ia, int ja, int *descA, scomplex *B, int ib, int jb, int *descB, int ctxt) {
    pcgemr2d_(&m,&n,A,&ia,&ja,descA,B,&ib,&jb,descB,&ctxt);
  }

  /* PxTRAN */
  template<typename T> inline void ptranc(int, int, T, T*, int, int, int *, T, T*, int, int, int *);
  template<> inline void ptranc<  double>(int m, int n,   double alpha,   double *A, int ia, int ja, int *descA,   double beta,   double *C, int ic, int jc, int *descC) {
    pdtran_(&m,&n,&alpha,A,&ia,&ja,descA,&beta,C,&ic,&jc,descC);
  }
  template<> inline void ptranc<   float>(int m, int n,    float alpha,    float *A, int ia, int ja, int *descA,    float beta,    float *C, int ic, int jc, int *descC) {
    pstran_(&m,&n,&alpha,A,&ia,&ja,descA,&beta,C,&ic,&jc,descC);
  }
  template<> inline void ptranc<dcomplex>(int m, int n, dcomplex alpha, dcomplex *A, int ia, int ja, int *descA, dcomplex beta, dcomplex *C, int ic, int jc, int *descC) {
    pztranc_(&m,&n,&alpha,A,&ia,&ja,descA,&beta,C,&ic,&jc,descC);
  }
  template<> inline void ptranc<scomplex>(int m, int n, scomplex alpha, scomplex *A, int ia, int ja, int *descA, scomplex beta, scomplex *C, int ic, int jc, int *descC) {
    pctranc_(&m,&n,&alpha,A,&ia,&ja,descA,&beta,C,&ic,&jc,descC);
  }

  /* PxGEQPF; this one is used only in the examples */
  template<typename T> inline void pgeqpf (int, int, T *A, int, int , int *, int, int, int, int, int, int);
  template<> inline void pgeqpf< double>(int m, int n,   double *A, int ia, int ja, int *descA, int mb, int nb, int myrow, int mycol, int nprow, int npcol) {
    int *ipiv;
    int locrA, loccA, lwork, ierr;
    double *tau, *work;
    locrA=numroc_(&m,&mb,&myrow,&IZERO,&nprow);
    loccA=numroc_(&n,&nb,&mycol,&IZERO,&npcol);
    ipiv=NULL;
    ipiv=new int[loccA];
    tau=NULL;
    tau=new double[loccA];
    lwork=3*(1+locrA+loccA);
    work=new double[lwork];
    pdgeqpf_(&m,&n,A,&ia,&ja,descA,ipiv,tau,work,&lwork,&ierr);
    delete[] tau;
    delete[] work;
    delete[] ipiv;
  }
  template<> inline void pgeqpf<  float>(int m, int n,    float *A, int ia, int ja, int *descA, int mb, int nb, int myrow, int mycol, int nprow, int npcol) {
    int *ipiv;
    int locrA, loccA, lwork, ierr;
    float *tau, *work;
    locrA=numroc_(&m,&mb,&myrow,&IZERO,&nprow);
    loccA=numroc_(&n,&nb,&mycol,&IZERO,&npcol);
    ipiv=NULL;
    ipiv=new int[loccA];
    tau=NULL;
    tau=new float[loccA];
    lwork=3*(1+locrA+loccA);
    work=new float[lwork];
    psgeqpf_(&m,&n,A,&ia,&ja,descA,ipiv,tau,work,&lwork,&ierr);
    delete[] tau;
    delete[] work;
    delete[] ipiv;
  }
  template<> inline void pgeqpf<dcomplex>(int m, int n, dcomplex *A, int ia, int ja, int *descA, int mb, int nb, int myrow, int mycol, int nprow, int npcol) {
    int *ipiv;
    int locrA, loccA, lwork, lrwork, ierr;
    dcomplex *tau, *work;
    double *rwork;
    locrA=numroc_(&m,&mb,&myrow,&IZERO,&nprow);
    loccA=numroc_(&n,&nb,&mycol,&IZERO,&npcol);
    ipiv=NULL;
    ipiv=new int[loccA];
    tau=NULL;
    tau=new dcomplex[loccA];
    lwork=3*(1+locrA+loccA);
    work=new dcomplex[lwork];
    lrwork=2*(1+loccA);
    rwork= new double[lrwork];
    pzgeqpf_(&m,&n,A,&ia,&ja,descA,ipiv,tau,work,&lwork,rwork,&lrwork,&ierr);
    delete[] tau;
    delete[] work;
    delete[] ipiv;
    delete[] rwork;
  }
  template<> inline void pgeqpf<scomplex>(int m, int n, scomplex *A, int ia, int ja, int *descA, int mb, int nb, int myrow, int mycol, int nprow, int npcol) {
    int *ipiv;
    int locrA, loccA, lwork, lrwork, ierr;
    scomplex *tau, *work;
    float *rwork;
    locrA=numroc_(&m,&mb,&myrow,&IZERO,&nprow);
    loccA=numroc_(&n,&nb,&mycol,&IZERO,&npcol);
    ipiv=NULL;
    ipiv=new int[loccA];
    tau=NULL;
    tau=new scomplex[loccA];
    lwork=3*(1+locrA+loccA);
    work=new scomplex[lwork];
    lrwork=2*(1+loccA);
    rwork= new float[lrwork];
    pcgeqpf_(&m,&n,A,&ia,&ja,descA,ipiv,tau,work,&lwork,rwork,&lrwork,&ierr);
    delete[] tau;
    delete[] work;
    delete[] ipiv;
    delete[] rwork;
  }

  /* PxGEQPFmod */
  template<typename T, typename S> inline void pgeqpfmod(int, int, T *, int, int , int *, int *, int *, int *, S, int, int, int, int, int, int);
  template<> inline void pgeqpfmod(int m, int n, double *A, int ia, int ja, int *descA, int *J, int *piv, int *r, double tol, int mb, int nb, int myrow, int mycol, int nprow, int npcol) {
    double *tau, *twork;
    int *ipiv;
    int lwork, info, locrA, loccA;

    locrA=numroc_(&m,&mb,&myrow,&IZERO,&nprow);
    loccA=numroc_(&n,&nb,&mycol,&IZERO,&npcol);
    lwork=3*(1+locrA+loccA);
    twork=new double[lwork];
    tau=new double[loccA];
    ipiv=new int[n];
    pdgeqpfmod_(&m,&n,A,&IONE,&IONE,descA,ipiv,tau,twork,&lwork,&info,J,piv,r,&tol);

    delete[] tau;
    delete[] twork;
  }
  template<> inline void pgeqpfmod(int m, int n,  float *A, int ia, int ja, int *descA, int *J, int *piv, int *r,  float tol, int mb, int nb, int myrow, int mycol, int nprow, int npcol) {
    float *tau, *twork;
    int *ipiv;
    int lwork, info, locrA, loccA;

    locrA=numroc_(&m,&mb,&myrow,&IZERO,&nprow);
    loccA=numroc_(&n,&nb,&mycol,&IZERO,&npcol);
    lwork=3*(1+locrA+loccA);
    twork=new float[lwork];
    tau=new float[loccA];
    ipiv=new int[n];
    psgeqpfmod_(&m,&n,A,&IONE,&IONE,descA,ipiv,tau,twork,&lwork,&info,J,piv,r,&tol);

    delete[] tau;
    delete[] twork;
  }
  template<> inline void pgeqpfmod(int m, int n, dcomplex *A, int ia, int ja, int *descA, int *J, int *piv, int *r, double tol, int mb, int nb, int myrow, int mycol, int nprow, int npcol) {
    dcomplex *tau, *twork;
    int *ipiv;
    double *rwork;
    int lwork, lrwork, info, locrA, loccA;
    locrA=numroc_(&m,&mb,&myrow,&IZERO,&nprow);
    loccA=numroc_(&n,&nb,&mycol,&IZERO,&npcol);
    lwork=3*(1+locrA+loccA);
    twork=new dcomplex[lwork];
    lrwork=2*(1+loccA);
    rwork= new double[lrwork];
    tau=new dcomplex[loccA];
    ipiv=new int[n];
    pzgeqpfmod_(&m,&n,A,&IONE,&IONE,descA,ipiv,tau,twork,&lwork,rwork,&lrwork,&info,J,piv,r,&tol);

    delete[] tau;
    delete[] twork;
    delete[] rwork;
  }
   template<> inline void pgeqpfmod(int m, int n, scomplex *A, int ia, int ja, int *descA, int *J, int *piv, int *r,  float tol, int mb, int nb, int myrow, int mycol, int nprow, int npcol) {
    scomplex *tau, *twork;
    int *ipiv;
    float *rwork;
    int lwork, lrwork, info, locrA, loccA;
    locrA=numroc_(&m,&mb,&myrow,&IZERO,&nprow);
    loccA=numroc_(&n,&nb,&mycol,&IZERO,&npcol);
    lwork=3*(1+locrA+loccA);
    twork=new scomplex[lwork];
    lrwork=2*(1+loccA);
    rwork= new float[lrwork];
    tau=new scomplex[loccA];
    ipiv=new int[n];
    pcgeqpfmod_(&m,&n,A,&IONE,&IONE,descA,ipiv,tau,twork,&lwork,rwork,&lrwork,&info,J,piv,r,&tol);

    delete[] tau;
    delete[] twork;
    delete[] rwork;
  }

  /* PxGETRF */
  template<typename T> inline int pgetrf(int, int, T *, int, int, int *, int *);
  template<> inline int pgetrf(int m, int n,   double *A, int ia, int ja, int *descA, int *ipiv) {
    int info;
    pdgetrf_(&m,&n,A,&ia,&ja,descA,ipiv,&info);
    return info;
  }
  template<> inline int pgetrf(int m, int n,    float *A, int ia, int ja, int *descA, int *ipiv) {
    int info;
    psgetrf_(&m,&n,A,&ia,&ja,descA,ipiv,&info);
    return info;
  }
  template<> inline int pgetrf(int m, int n, dcomplex *A, int ia, int ja, int *descA, int *ipiv) {
    int info;
    pzgetrf_(&m,&n,A,&ia,&ja,descA,ipiv,&info);
    return info;
  }
  template<> inline int pgetrf(int m, int n, scomplex *A, int ia, int ja, int *descA, int *ipiv) {
    int info;
    pcgetrf_(&m,&n,A,&ia,&ja,descA,ipiv,&info);
    return info;
  }

  /* PxGETRS */
  template<typename T> inline int pgetrs(char, int, int, T *, int, int, int *, int *, T *, int, int, int *);
  template<> inline int pgetrs(char trans, int m, int n,   double *A, int ia, int ja, int *descA, int *ipiv, double *B, int ib, int jb, int *descB) {
    int info;
    pdgetrs_(&trans,&m,&n,A,&ia,&ja,descA,ipiv,B,&ib,&jb,descB,&info);
    return info;
  }
  template<> inline int pgetrs(char trans, int m, int n,    float *A, int ia, int ja, int *descA, int *ipiv, float *B, int ib, int jb, int *descB) {
    int info;
    psgetrs_(&trans,&m,&n,A,&ia,&ja,descA,ipiv,B,&ib,&jb,descB,&info);
    return info;
  }
  template<> inline int pgetrs(char trans, int m, int n, dcomplex *A, int ia, int ja, int *descA, int *ipiv, dcomplex *B, int ib, int jb, int *descB) {
    int info;
    pzgetrs_(&trans,&m,&n,A,&ia,&ja,descA,ipiv,B,&ib,&jb,descB,&info);
    return info;
  }
  template<> inline int pgetrs(char trans, int m, int n, scomplex *A, int ia, int ja, int *descA, int *ipiv, scomplex *B, int ib, int jb, int *descB) {
    int info;
    pcgetrs_(&trans,&m,&n,A,&ia,&ja,descA,ipiv,B,&ib,&jb,descB,&info);
    return info;
  }

  /* PxGELQF */
  template<typename T> inline void pgelqf (int, int, T *A, int, int , int *, T *, int, int, int, int, int, int);
  template<> inline void pgelqf< double>(int m, int n,   double *A, int ia, int ja, int *descA, double * tau, int mb, int nb, int myrow, int mycol, int nprow, int npcol) {
    int locrA, loccA, lwork, ierr;
    double *work;
    locrA=numroc_(&m,&mb,&myrow,&IZERO,&nprow);
    loccA=numroc_(&n,&nb,&mycol,&IZERO,&npcol);
    lwork=mb*(mb+locrA+loccA);
    work=new double[lwork];
    pdgelqf_(&m,&n,A,&ia,&ja,descA,tau,work,&lwork,&ierr);
    delete[] work;
  }
  template<> inline void pgelqf<  float>(int m, int n,    float *A, int ia, int ja, int *descA, float *tau, int mb, int nb, int myrow, int mycol, int nprow, int npcol) {
    int locrA, loccA, lwork, ierr;
    float *work;
    locrA=numroc_(&m,&mb,&myrow,&IZERO,&nprow);
    loccA=numroc_(&n,&nb,&mycol,&IZERO,&npcol);
    lwork=mb*(mb+locrA+loccA);
    work=new float[lwork];
    psgelqf_(&m,&n,A,&ia,&ja,descA,tau,work,&lwork,&ierr);
    delete[] work;
  }
  template<> inline void pgelqf<dcomplex>(int m, int n, dcomplex *A, int ia, int ja, int *descA, dcomplex *tau, int mb, int nb, int myrow, int mycol, int nprow, int npcol) {
    int locrA, loccA, lwork, ierr;
    dcomplex *work;
    locrA=numroc_(&m,&mb,&myrow,&IZERO,&nprow);
    loccA=numroc_(&n,&nb,&mycol,&IZERO,&npcol);
    lwork=mb*(mb+locrA+loccA);
    work=new dcomplex[lwork];
    pzgelqf_(&m,&n,A,&ia,&ja,descA,tau,work,&lwork,&ierr);
    delete[] work;
  }
  template<> inline void pgelqf<scomplex>(int m, int n, scomplex *A, int ia, int ja, int *descA, scomplex *tau, int mb, int nb, int myrow, int mycol, int nprow, int npcol) {
    int locrA, loccA, lwork, ierr;
    scomplex *work;
    locrA=numroc_(&m,&mb,&myrow,&IZERO,&nprow);
    loccA=numroc_(&n,&nb,&mycol,&IZERO,&npcol);
    lwork=mb*(mb+locrA+loccA);
    work=new scomplex[lwork];
    pcgelqf_(&m,&n,A,&ia,&ja,descA,tau,work,&lwork,&ierr);
    delete[] work;
  }

  /* PxxxGLQ */
  template<typename T> inline void pxxglq(int, int, int, T *, int, int, int *, T *, int, int, int, int, int, int);
  template<> inline void pxxglq<double>(int m, int n, int k, double *A, int ia, int ja, int *descA, double *tau, int mb, int nb, int myrow, int mycol, int nprow, int npcol) {
    int locrA, loccA, lwork, ierr;
    double *work;
    locrA=numroc_(&m,&mb,&myrow,&IZERO,&nprow);
    loccA=numroc_(&n,&nb,&mycol,&IZERO,&npcol);
    lwork=mb*(mb+locrA+loccA);
    work=new double[lwork];
    pdorglq_(&m,&n,&k,A,&ia,&ja,descA,tau,work,&lwork,&ierr);
    delete[] work;
  }
  template<> inline void pxxglq<  float>(int m, int n, int k, float *A, int ia, int ja, int *descA, float *tau, int mb, int nb, int myrow, int mycol, int nprow, int npcol) {
    int locrA, loccA, lwork, ierr;
    float *work;
    locrA=numroc_(&m,&mb,&myrow,&IZERO,&nprow);
    loccA=numroc_(&n,&nb,&mycol,&IZERO,&npcol);
    lwork=mb*(mb+locrA+loccA);
    work=new float[lwork];
    psorglq_(&m,&n,&k,A,&ia,&ja,descA,tau,work,&lwork,&ierr);
    delete[] work;
  }
  template<> inline void pxxglq<dcomplex>(int m, int n, int k, dcomplex *A, int ia, int ja, int *descA, dcomplex *tau, int mb, int nb, int myrow, int mycol, int nprow, int npcol) {
    int locrA, loccA, lwork, ierr;
    dcomplex *work;
    locrA=numroc_(&m,&mb,&myrow,&IZERO,&nprow);
    loccA=numroc_(&n,&nb,&mycol,&IZERO,&npcol);
    lwork=mb*(mb+locrA+loccA);
    work=new dcomplex[lwork];
    pzunglq_(&m,&n,&k,A,&ia,&ja,descA,tau,work,&lwork,&ierr);
    delete[] work;
  }
  template<> inline void pxxglq<scomplex>(int m, int n, int k, scomplex *A, int ia, int ja, int *descA, scomplex *tau, int mb, int nb, int myrow, int mycol, int nprow, int npcol) {
    int locrA, loccA, lwork, ierr;
    scomplex *work;
    locrA=numroc_(&m,&mb,&myrow,&IZERO,&nprow);
    loccA=numroc_(&n,&nb,&mycol,&IZERO,&npcol);
    lwork=mb*(mb+locrA+loccA);
    work=new scomplex[lwork];
    pcunglq_(&m,&n,&k,A,&ia,&ja,descA,tau,work,&lwork,&ierr);
    delete[] work;
  }

  /* PxNRM2; this one is used only in the examples */
  template<typename T, typename S> inline S pnrm2(int, T *, int, int, int *, int);
  template<> inline double pnrm2<double, double>(int n, double *X, int ix, int jx, int *descX, int incx) {
    double nrm;
    pdnrm2_(&n,&nrm,X,&ix,&jx,descX,&incx);
    return nrm;
  }
  template<> inline float pnrm2<float, float>(int n, float *X, int ix, int jx, int *descX, int incx) {
    float nrm;
    psnrm2_(&n,&nrm,X,&ix,&jx,descX,&incx);
    return nrm;
  }
  template<> inline double pnrm2<dcomplex, double>(int n, dcomplex *X, int ix, int jx, int *descX, int incx) {
    double nrm;
    pdznrm2_(&n,&nrm,X,&ix,&jx,descX,&incx);
    return nrm;
  }
  template<> inline float pnrm2<scomplex, float>(int n, scomplex *X, int ix, int jx, int *descX, int incx) {
    float nrm;
    pscnrm2_(&n,&nrm,X,&ix,&jx,descX,&incx);
    return nrm;
  }

#endif // SCALAPACK_HPP
